import pytest
import xarray as xr

from climpred import HindcastEnsemble, set_options
from climpred.exceptions import DimensionError
from climpred.options import OPTIONS


def test_hindcastEnsemble_init(hind_ds_initialized_1d):
    """Test to see hindcast ensemble can be initialized with xr.Dataset."""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    assert hindcast


def test_hindcastEnsemble_init_da(hind_da_initialized_1d):
    """Test to see hindcast ensemble can be initialized with xr.DataArray."""
    hindcast = HindcastEnsemble(hind_da_initialized_1d)
    assert hindcast


def test_add_observations(hind_ds_initialized_1d, reconstruction_ds_1d):
    """Test to see if observations can be added to the HindcastEnsemble"""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    hindcast = hindcast.add_observations(reconstruction_ds_1d)
    assert hindcast.get_observations()


def test_add_observations_da_1d(hind_ds_initialized_1d, observations_da_1d):
    """Test to see if observations can be added to the HindcastEnsemble as a da"""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    hindcast = hindcast.add_observations(observations_da_1d)
    assert hindcast.get_observations()


def test_add_uninitialized(hind_ds_initialized_1d, hist_ds_uninitialized_1d):
    """Test to see if an uninitialized ensemble can be added to the HindcastEnsemble"""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    hindcast = hindcast.add_uninitialized(hist_ds_uninitialized_1d)
    assert hindcast.get_uninitialized()


def test_add_hist_da_uninitialized_1d(hind_ds_initialized_1d, hist_da_uninitialized_1d):
    """Test to see if da uninitialized ensemble can be added to the HindcastEnsemble"""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    hindcast = hindcast.add_uninitialized(hist_da_uninitialized_1d)
    assert hindcast.get_uninitialized()


def test_verify(hindcast_hist_obs_1d):
    """Test to see if HindcastEnsemble.verify() works."""
    hindcast = hindcast_hist_obs_1d
    hindcast.verify(metric="acc", comparison="e2o", dim="init", alignment="same_verif")


def test_isel_xarray_func(hindcast_hist_obs_1d):
    """Test whether applying isel to the objects works."""
    hindcast = hindcast_hist_obs_1d
    hindcast = hindcast.isel(lead=0, init=slice(0, 3)).isel(time=slice(5, 10))
    assert hindcast.get_initialized().init.size == 3
    assert hindcast.get_initialized().lead.size == 1
    assert hindcast.get_observations().time.size == 5


def test_get_initialized(hindcast_hist_obs_1d):
    """Test whether get_initialized method works."""
    assert hindcast_hist_obs_1d.get_initialized().identical(
        hindcast_hist_obs_1d._datasets["initialized"]
    )


def test_get_uninitialized(hindcast_hist_obs_1d):
    """Test whether get_uninitialized method works."""
    assert hindcast_hist_obs_1d.get_uninitialized().identical(
        hindcast_hist_obs_1d._datasets["uninitialized"]
    )


def test_get_observations(hindcast_hist_obs_1d):
    """Tests whether get_observations method works."""
    assert hindcast_hist_obs_1d.get_observations().identical(
        hindcast_hist_obs_1d._datasets["observations"]
    )


def test_inplace(
    hind_ds_initialized_1d, reconstruction_ds_1d, hist_ds_uninitialized_1d
):
    """Tests that inplace operations do not work."""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    # Adding observations.
    hindcast.add_observations(reconstruction_ds_1d)
    with_obs = hindcast.add_observations(reconstruction_ds_1d)
    assert hindcast != with_obs
    # Adding an uninitialized ensemble.
    hindcast.add_uninitialized(hist_ds_uninitialized_1d)
    with_uninit = hindcast.add_uninitialized(hist_ds_uninitialized_1d)
    assert hindcast != with_uninit
    # Applying arbitrary func.
    hindcast.sum("init")
    summed = hindcast.sum("init")
    assert hindcast != summed


@pytest.mark.parametrize("call", ["verify", "bootstrap"])
@pytest.mark.parametrize(
    "dim",
    [set(["init"]), list(["init"]), tuple(["init"]), "init"],
    ids=["set", "list", "tuple", "str"],
)
def test_dim_input_type(hindcast_hist_obs_1d, dim, call):
    """Test verify and bootstrap for different dim types."""
    kw = dict(iterations=2) if call == "bootstrap" else {}
    assert getattr(hindcast_hist_obs_1d.isel(lead=range(3)), call)(
        metric="rmse", comparison="e2o", dim=dim, alignment="same_verifs", **kw
    )


def test_verify_metric_kwargs(hindcast_hist_obs_1d):
    """Test that HindcastEnsemble works with metrics using metric_kwargs."""
    assert hindcast_hist_obs_1d.verify(
        metric="threshold_brier_score",
        comparison="m2o",
        dim="member",
        threshold=0.5,
        reference="uninitialized",
        alignment="same_verifs",
    )


def test_verify_fails_expected_metric_kwargs(hindcast_hist_obs_1d):
    """Test that HindcastEnsemble fails when metric_kwargs expected but not given."""
    hindcast = hindcast_hist_obs_1d
    with pytest.raises(ValueError, match="Please provide threshold"):
        hindcast.verify(
            metric="threshold_brier_score",
            comparison="m2o",
            dim="member",
            alignment="same_verifs",
        )


def test_calendar_matching_observations(hind_ds_initialized_1d, reconstruction_ds_1d):
    """Tests that error is thrown if calendars mismatch when adding observations."""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    reconstruction_ds_1d["time"] = xr.cftime_range(
        start="1950",
        periods=reconstruction_ds_1d.time.size,
        freq="MS",
        calendar="all_leap",
    )
    with pytest.raises(ValueError, match="does not match"):
        hindcast.add_observations(reconstruction_ds_1d)


def test_calendar_matching_uninitialized(
    hind_ds_initialized_1d, hist_ds_uninitialized_1d
):
    """Tests that error is thrown if calendars mismatch when adding uninitialized."""
    hindcast = HindcastEnsemble(hind_ds_initialized_1d)
    hist_ds_uninitialized_1d["time"] = xr.cftime_range(
        start="1950",
        periods=hist_ds_uninitialized_1d.time.size,
        freq="MS",
        calendar="all_leap",
    )
    with pytest.raises(ValueError, match="does not match"):
        hindcast.add_uninitialized(hist_ds_uninitialized_1d)


@pytest.mark.parametrize("metric", ["mse", "crps"])
def test_verify_reference_same_dims(hindcast_hist_obs_1d, metric):
    """Test that verify returns the same dimensionality regardless of reference."""
    hindcast = hindcast_hist_obs_1d.isel(lead=range(3), init=range(10))
    if metric == "mse":
        comparison = "e2o"
        dim = "init"
    elif metric == "crps":
        comparison = "m2o"
        dim = ["member", "init"]
    alignment = "same_verif"
    actual_no_ref = hindcast.verify(
        metric=metric,
        comparison=comparison,
        dim=dim,
        alignment=alignment,
        reference=None,
    )
    actual_uninit_ref = hindcast.verify(
        metric=metric,
        comparison=comparison,
        dim=dim,
        alignment=alignment,
        reference="uninitialized",
    )
    actual_pers_ref = hindcast.verify(
        metric=metric,
        comparison=comparison,
        dim=dim,
        alignment=alignment,
        reference="persistence",
    )
    actual_clim_ref = hindcast.verify(
        metric=metric,
        comparison=comparison,
        dim=dim,
        alignment=alignment,
        reference="climatology",
    )
    assert actual_uninit_ref.skill.size == 2
    assert actual_pers_ref.skill.size == 2
    assert actual_clim_ref.skill.size == 2
    # no additional dimension, +1 because initialized squeezed
    assert len(actual_no_ref.dims) + 1 == len(actual_pers_ref.dims)
    assert len(actual_no_ref.dims) + 1 == len(actual_uninit_ref.dims)
    assert len(actual_no_ref.dims) + 1 == len(actual_clim_ref.dims)
    assert len(actual_pers_ref.dims) == len(actual_uninit_ref.dims)
    assert len(actual_pers_ref.dims) == len(actual_clim_ref.dims)


def test_HindcastEnsemble_multidim_initialized_lessdim_verif(hindcast_hist_obs_1d):
    """Test to see if HindcastEnsemble allow broadcast over dimensions in initialized but not in observations, e.g. model dim which is not available in observations."""
    initialized = hindcast_hist_obs_1d.get_initialized()
    obs = hindcast_hist_obs_1d.get_observations()
    hind = HindcastEnsemble(
        initialized.expand_dims("model").isel(model=[0] * 2)
    ).add_observations(obs)
    skill = hind.verify(
        metric="acc", dim="init", comparison="e2o", alignment="same_inits"
    )
    assert "model" in skill.dims


def test_HindcastEnsemble_multidim_verif_lessdim_initialized(hindcast_hist_obs_1d):
    """Test if HindcastEnsemble initialization fails if observations contains more dims than initialized."""
    initialized = hindcast_hist_obs_1d.get_initialized()
    obs = hindcast_hist_obs_1d.get_observations()
    with pytest.raises(
        DimensionError, match="Verification contains more dimensions than initialized"
    ):
        HindcastEnsemble(initialized).add_observations(
            obs.expand_dims("model").isel(model=[0] * 2)
        )


@pytest.mark.parametrize(
    "dim,new_dim,cf_standard_name",
    [
        ("init", "forecast_time", "forecast_reference_time"),
        ("lead", "lead_time", "forecast_period"),
        ("member", "number", "realization"),
    ],
)
def test_HindcastEnsemble_instantiating_standard_name(
    da_lead, dim, new_dim, cf_standard_name
):
    """Instantiating a HindcastEnsemble without init dim only works if matching CF standard name is set."""
    init = (
        da_lead.to_dataset(name="var").expand_dims("member").assign_coords(member=[1])
    )
    init["init"] = xr.cftime_range(start="2000", periods=init.init.size, freq="YS")
    init["lead"].attrs["units"] = "years"
    # change to non CLIMPRED_DIMS
    init = init.rename({dim: new_dim})

    if dim != "member":  # member not required
        with pytest.raises(
            DimensionError,
            match="Your PredictionEnsemble object must contain the following dimensions",
        ):
            HindcastEnsemble(init)

    init[new_dim].attrs["standard_name"] = cf_standard_name
    # find renamed after warning
    with pytest.warns(UserWarning, match="but renamed dimension"):
        init = HindcastEnsemble(init).get_initialized()
        assert dim in init.dims, print(init.dims, init.coords)



import logging

from climpred.prediction import compute_hindcast


def test_log_compute_hindcast(
    hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime, caplog
):
    """Tests that logging works for compute_hindcast."""
    LOG_STRINGS = ["lead", "inits", "verifs"]
    with caplog.at_level(logging.INFO):
        compute_hindcast(hind_ds_initialized_1d_cftime, reconstruction_ds_1d_cftime)
        for i, record in enumerate(caplog.record_tuples):
            # Skip header information.
            if i >= 2:
                print(record)
                assert all(x in record[2] for x in LOG_STRINGS)


def test_log_HindcastEnsemble_verify(hindcast_hist_obs_1d, caplog):
    """Test that verify logs."""
    LOG_STRINGS = ["lead", "inits", "verifs"]
    with caplog.at_level(logging.INFO):
        hindcast_hist_obs_1d.verify(
            metric="mse", comparison="e2o", dim="init", alignment="same_verif"
        )
        for i, record in enumerate(caplog.record_tuples):
            # Skip header information.
            if i >= 2:
                print(record)
                assert all(x in record[2] for x in LOG_STRINGS)
                assert "initialized" in record[2]



import pytest
import xarray as xr

from climpred.stats import rm_poly
from climpred.testing import assert_PredictionEnsemble

xr.set_options(display_style="text")


def test_PredictionEnsemble_raises_error(hindcast_hist_obs_1d):
    """Tests that PredictionEnsemble raises error."""
    with pytest.raises(TypeError):
        hindcast_hist_obs_1d.smooth({"lead": 2.5})  # expects int


def test_PredictionEnsemble_raises_warning(hindcast_hist_obs_1d):
    """Tests that PredictionEnsemble raises warning."""
    with pytest.warns(UserWarning):
        hindcast_hist_obs_1d.map(rm_poly, dim="init", deg=2)


def test_PredictionEnsemble_xr_calls(hindcast_hist_obs_1d):
    """Tests that PredictionEnsemble passes through xarray calls."""
    assert (
        hindcast_hist_obs_1d.isel(lead=[1, 2])
        .get_initialized()
        .identical(hindcast_hist_obs_1d.get_initialized().isel(lead=[1, 2]))
    )


def test_PredictionEnsemble_map_dim_or(hindcast_hist_obs_1d):
    """Tests that PredictionEnsemble allows dim0_or_dim1 as kwargs without UserWarning."""
    with pytest.warns(None):  # no warnings
        he_or = hindcast_hist_obs_1d.map(rm_poly, dim="init_or_time", deg=2)
    assert he_or != hindcast_hist_obs_1d

    with pytest.warns(UserWarning) as record:  # triggers warnings
        he_chained = hindcast_hist_obs_1d.map(rm_poly, dim="init", deg=2).map(
            rm_poly, dim="time", deg=2
        )
    assert he_chained != hindcast_hist_obs_1d

    assert len(record) == 3  # for init, uninit and obs
    for r in record:
        assert "Error due to " in str(r.message)

    assert_PredictionEnsemble(he_or, he_chained)


def test_PredictionEnsemble_map_dim_or_fails_if_both_dims_in_dataset(
    hindcast_hist_obs_1d,
):
    """Tests that PredictionEnsemble with dim0_or_dim1 as kwargs fails if both dims in any dataset."""
    with pytest.raises(ValueError, match="cannot be both in"):
        hindcast_hist_obs_1d.map(rm_poly, dim="init_or_lead", deg=2)



import cftime
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from xarray.testing import assert_allclose

from climpred.bootstrap import bootstrap_perfect_model
from climpred.comparisons import PM_COMPARISONS, __m2c
from climpred.metrics import DETERMINISTIC_PM_METRICS, __pearson_r
from climpred.prediction import compute_hindcast, compute_perfect_model
from climpred.tutorial import load_dataset
from climpred.utils import (
    convert_time_index,
    find_start_dates_for_given_init,
    get_comparison_class,
    get_metric_class,
    intersect,
    lead_units_equal_control_time_stride,
    return_time_series_freq,
    shift_cftime_index,
    shift_cftime_singular,
)


def test_get_metric_class():
    """Test if passing in a string gets the right metric function."""
    actual = get_metric_class("pearson_r", DETERMINISTIC_PM_METRICS).name
    expected = __pearson_r.name
    assert actual == expected


def test_get_metric_class_fail():
    """Test if passing something not in the dict raises the right error."""
    with pytest.raises(KeyError) as excinfo:
        get_metric_class("not_metric", DETERMINISTIC_PM_METRICS)
    assert "Specify metric from" in str(excinfo.value)


def test_get_comparison_class():
    """Test if passing in a string gets the right comparison function."""
    actual = get_comparison_class("m2c", PM_COMPARISONS).name
    expected = __m2c.name
    assert actual == expected


def test_get_comparison_class_fail():
    """Test if passing something not in the dict raises the right error."""
    with pytest.raises(KeyError) as excinfo:
        get_comparison_class("not_comparison", PM_COMPARISONS)
    assert "Specify comparison from" in str(excinfo.value)


def test_intersect():
    """Test if the intersect (overlap) of two lists work."""
    x = [1, 5, 6]
    y = [1, 6, 7]
    actual = intersect(x, y)
    expected = np.array([1, 6])
    assert all(a == e for a, e in zip(actual, expected))


def test_da_assign_attrs(PM_ds_initialized_1d, PM_ds_control_1d):
    """Test assigning attrs for compute_perfect_model and dataarrays."""
    metric = "pearson_r"
    comparison = "m2e"
    actual = compute_perfect_model(
        PM_ds_initialized_1d,
        PM_ds_control_1d,
        metric=metric,
        comparison=comparison,
    ).attrs
    assert actual["metric"] == metric
    assert actual["comparison"] == comparison
    if metric == "pearson_r":
        assert actual["units"] == "None"
    assert (
        actual["prediction_skill"]
        == "calculated by climpred https://climpred.readthedocs.io/"
    )


def test_ds_assign_attrs(PM_ds_initialized_1d, PM_ds_control_1d):
    """Test assigning attrs for datasets."""
    metric = "mse"
    comparison = "m2e"
    dim = ["init", "member"]
    PM_ds_initialized_1d.attrs["units"] = "C"
    actual = compute_perfect_model(
        PM_ds_initialized_1d,
        PM_ds_control_1d,
        metric=metric,
        comparison=comparison,
        dim=dim,
    ).attrs
    assert actual["metric"] == metric
    assert actual["comparison"] == comparison
    if metric == "pearson_r":
        assert actual["units"] == "None"
    assert actual["units"] == "(C)^2"
    assert actual["dim"] == dim


def test_bootstrap_pm_assign_attrs():
    """Test assigning attrs for bootstrap_perfect_model."""
    v = "tos"
    metric = "pearson_r"
    comparison = "m2e"
    ITERATIONS = 3
    sig = 95
    da = load_dataset("MPI-PM-DP-1D")[v].isel(area=1, period=-1)
    control = load_dataset("MPI-control-1D")[v].isel(area=1, period=-1)
    actual = bootstrap_perfect_model(
        da,
        control,
        metric=metric,
        comparison=comparison,
        iterations=ITERATIONS,
        sig=sig,
    ).attrs
    assert actual["metric"] == metric
    assert actual["comparison"] == comparison
    assert actual["bootstrap_iterations"] == ITERATIONS
    assert str(round((1 - sig / 100) / 2, 3)) in actual["confidence_interval_levels"]
    if metric == "pearson_r":
        assert actual["units"] == "None"


def test_hindcast_assign_attrs():
    """Test assigning attrs for compute_hindcast."""
    metric = "pearson_r"
    comparison = "e2o"
    da = load_dataset("CESM-DP-SST")
    control = load_dataset("ERSST")
    actual = compute_hindcast(da, control, metric=metric, comparison=comparison).attrs
    assert actual["metric"] == metric
    assert actual["comparison"] == comparison
    if metric == "pearson_r":
        assert actual["units"] == "None"


def test_cftime_index_unchanged():
    """Tests that a CFTime index going through convert time is unchanged."""
    inits = xr.cftime_range("1990", "2000", freq="Y", calendar="noleap")
    da = xr.DataArray(np.random.rand(len(inits)), dims="init", coords=[inits])
    new_inits = convert_time_index(da, "init", "")
    assert_allclose(new_inits.init, da.init)


def test_pandas_datetime_converted_to_cftime():
    """Tests that a pd.DatetimeIndex is converted to xr.CFTimeIndex."""
    inits = pd.date_range("1990", "2000", freq="YS")
    da = xr.DataArray(np.random.rand(len(inits)), dims="init", coords=[inits])
    new_inits = convert_time_index(da, "init", "")
    assert isinstance(new_inits["init"].to_index(), xr.CFTimeIndex)


def test_int64_converted_to_cftime():
    """Tests the xr.Int64Index is converted to xr.CFTimeIndex."""
    inits = np.arange(1990, 2000)
    da = xr.DataArray(np.random.rand(len(inits)), dims="init", coords=[inits])
    new_inits = convert_time_index(da, "init", "")
    assert isinstance(new_inits["init"].to_index(), xr.CFTimeIndex)


def test_float64_converted_to_cftime():
    """Tests the xr.Float64Index is converted to xr.CFTimeIndex."""
    inits = np.arange(1990, 2000) * 1.0
    da = xr.DataArray(np.random.rand(len(inits)), dims="init", coords=[inits])
    new_inits = convert_time_index(da, "init", "")
    assert isinstance(new_inits["init"].to_index(), xr.CFTimeIndex)


def test_numeric_index_auto_appends_lead_attrs():
    """Tests that for numeric inits, lead units are automatically set to 'years'"""
    lead = np.arange(3)
    int_inits = np.arange(1990, 2000)
    float_inits = int_inits * 1.0
    int_da = xr.DataArray(
        np.random.rand(len(int_inits), len(lead)),
        dims=["init", "lead"],
        coords=[int_inits, lead],
    )
    float_da = xr.DataArray(
        np.random.rand(len(float_inits), len(lead)),
        dims=["init", "lead"],
        coords=[float_inits, lead],
    )
    new_int_da = convert_time_index(int_da, "init", "")
    new_float_da = convert_time_index(float_da, "init", "")
    assert new_int_da.lead.attrs["units"] == "years"
    assert new_float_da.lead.attrs["units"] == "years"


def test_convert_time_index_does_not_overwrite():
    """Tests that `convert_time_index` does not overwrite the original index."""
    inits = np.arange(1990, 2000)
    da = xr.DataArray(np.random.rand(len(inits)), dims="init", coords=[inits])
    new_inits = convert_time_index(da, "init", "")
    assert isinstance(da.init.to_index(), pd.Int64Index)
    assert isinstance(new_inits.init.to_index(), xr.CFTimeIndex)


def test_irregular_initialization_dates():
    """Tests that irregularly spaced initializations convert properly."""
    inits = np.arange(1990, 2010)
    inits = np.delete(inits, [3, 5, 8, 12, 15])
    da = xr.DataArray(np.random.rand(len(inits)), dims="init", coords=[inits])
    new_inits = convert_time_index(da, "init", "")
    assert (new_inits["init"].to_index().year == inits).all()


def test_shift_cftime_singular():
    """Tests that a singular ``cftime`` is shifted the appropriate amount."""
    cftime_initial = cftime.DatetimeNoLeap(1990, 1, 1)
    cftime_expected = cftime.DatetimeNoLeap(1990, 3, 1)
    # Shift forward two months at month start.
    cftime_from_func = shift_cftime_singular(cftime_initial, 2, "MS")
    assert cftime_expected == cftime_from_func


def test_shift_cftime_index():
    """Tests that ``CFTimeIndex`` is shifted by the appropriate amount."""
    idx = xr.cftime_range("1990", "2000", freq="YS")
    da = xr.DataArray(np.random.rand(len(idx)), dims="time", coords=[idx])
    expected = idx.shift(3, "YS")
    res = shift_cftime_index(da, "time", 3, "YS")
    assert (expected == res).all()


@pytest.mark.parametrize(
    "init",
    [
        pytest.lazy_fixture("PM_ds_initialized_1d_ym_cftime"),
        pytest.lazy_fixture("PM_ds_initialized_1d_mm_cftime"),
        pytest.lazy_fixture("PM_ds_initialized_1d_dm_cftime"),
    ],
)
def test_return_time_series_freq_freq_init_pm(init):
    """Test that return_time_series_freq returns expected freq for different lead
    units."""
    actual = return_time_series_freq(init, "init")
    expected = init.lead.attrs["units"].strip("s")
    assert actual == expected


@pytest.mark.parametrize(
    "init, control",
    [
        (
            pytest.lazy_fixture("PM_ds_initialized_1d_ym_cftime"),
            pytest.lazy_fixture("PM_ds_control_1d_ym_cftime"),
        ),
        (
            pytest.lazy_fixture("PM_ds_initialized_1d_mm_cftime"),
            pytest.lazy_fixture("PM_ds_control_1d_mm_cftime"),
        ),
        (
            pytest.lazy_fixture("PM_ds_initialized_1d_dm_cftime"),
            pytest.lazy_fixture("PM_ds_control_1d_dm_cftime"),
        ),
    ],
)
def test_lead_units_equal_control_time_stride_freq(init, control):
    """Test that init_pm and control are compatible when both same freq."""
    assert lead_units_equal_control_time_stride(init, control)


def test_lead_units_equal_control_time_stride_daily_fails(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_dm_cftime
):
    """Test that init_pm annual and control daily is not compatible."""
    with pytest.raises(ValueError) as excinfo:
        lead_units_equal_control_time_stride(
            PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_dm_cftime
        )
        assert "Please provide the same temporal resolution for control.time" in str(
            excinfo.value
        )


def test_find_start_dates_for_given_init(
    PM_ds_initialized_1d_mm_cftime, PM_ds_control_1d_mm_cftime
):
    """Test that start dates are one year apart."""
    for init in PM_ds_initialized_1d_mm_cftime.init:
        start_dates = find_start_dates_for_given_init(PM_ds_control_1d_mm_cftime, init)
        freq = return_time_series_freq(PM_ds_initialized_1d_mm_cftime, "init")
        assert return_time_series_freq(start_dates, "time") == "year"
        assert (getattr(start_dates.time.dt, freq) == getattr(init.dt, freq)).all()
        # same number of start dates are years or one less
        assert start_dates.time.size - len(
            np.unique(PM_ds_control_1d_mm_cftime.time.dt.year.values)
        ) in [0, 1]



import numpy as np
import xarray as xr


def _last_item_cond_true(cond, dim):
    """Return the final item from cond that evaluates to True.

    Args:
        cond (xr.DataArray, xr.Dataset): User-defined boolean array where True means
            the system is predictable at the given lead. E.g., this could be based on
            the dynamical forecast beating a reference forecast, p values, confidence
            intervals, etc. cond should contain the dimension dim at the minimum.
        dim (str): Dimension to check for condition == True over.

    Returns:
        xr.DataArray, xr.Dataset: ``dim`` value until condition is True.

    """
    # force DataArray because isel (when transforming to dim space) requires DataArray
    if isinstance(cond, xr.Dataset):
        was_dataset = True
        cond = cond.to_array()
    else:
        was_dataset = False
    # index last True
    reached = cond.argmin(dim)
    # fix below one
    reached = reached.where(reached >= 1, np.nan)
    # reset where always true to len(lead)
    reached = reached.where(~cond.all("lead"), other=cond[dim].size)
    # fix locations where always nan to nan
    mask = cond.notnull().all("lead")  # ~(cond == False).all("lead")
    reached = reached.where(mask, other=np.nan)
    ## shift back into coordinate space ##
    # problem: cannot convert nan to idx in isel
    # therefore set to dim:0 and mask again afterwards
    reached_notnull = reached.notnull()  # remember where not masked
    reached = reached.where(
        reached.notnull(), other=cond.isel({dim: 0})
    )  # set nan to dim:0
    # take one index before calculated by argmin
    reached_dim_space = cond[dim].isel(
        {dim: reached.astype(int) - 1}
    )  # to not break conversion to dim space
    reached_dim_space = reached_dim_space.where(
        reached_notnull, other=np.nan
    )  # cleanup replace dim:0 with nan again
    if was_dataset:
        reached_dim_space = reached_dim_space.to_dataset(dim="variable").squeeze(
            drop=True
        )
        if "lead" in reached_dim_space.coords:
            reached_dim_space = reached_dim_space.drop("lead")
    return reached_dim_space


def horizon(cond):
    """Calculate the predictability horizon based on a condition ```cond``.

    Args:
        cond (xr.DataArray, xr.Dataset): User-defined boolean array where True means
            the system is predictable at the given lead. E.g., this could be based on
            the dynamical forecast beating a reference forecast, p values, confidence
            intervals, etc. cond should contain the dimension lead at the minimum.

    Returns:
        xr.DataArray, xr.Dataset: predictability horizon reduced by ``lead`` dimension.

    Example:
        >>> skill = PerfectModelEnsemble.verify(metric='acc', comparison='m2e',
        ...     dim=['init','member'], reference=['persistence'])
        >>> horizon(skill.sel(skill='initialized') >
        ...     skill.sel(skill='persistence'))
        <xarray.Dataset>
        Dimensions:  ()
        Data variables:
            tos      float64 15.0
        Attributes:
            units:          years
            standard_name:  forecast_period
            long_name:      Lead
            description:    Forecast period is the time interval between the forecast...


        >>> bskill = PerfectModelEnsemble.bootstrap(metric='acc', comparison='m2e',
        ...     dim=['init','member'], reference='uninitialized', iterations=201)
        >>> horizon(bskill.sel(skill='uninitialized', results='p') <= 0.05)
        <xarray.Dataset>
        Dimensions:  ()
        Coordinates:
            results  <U12 'p'
            skill    <U13 'uninitialized'
        Data variables:
            tos      float64 10.0
        Attributes:
            units:          years
            standard_name:  forecast_period
            long_name:      Lead
            description:    Forecast period is the time interval between the forecast...

    """
    ph = _last_item_cond_true(cond, "lead")
    if isinstance(ph, xr.DataArray):
        ph.attrs["units"] = cond["lead"].attrs["units"]
    elif isinstance(ph, xr.Dataset):
        for v in ph.data_vars:
            ph[v].attrs["units"] = cond["lead"].attrs["units"]
    return ph



import numpy as np
import pytest
import xarray as xr

from climpred.prediction import compute_perfect_model
from climpred.smoothing import (
    _reset_temporal_axis,
    _set_center_coord,
    smooth_goddard_2013,
    temporal_smoothing,
)
from climpred.testing import assert_PredictionEnsemble

try:
    from climpred.smoothing import spatial_smoothing_xesmf

    xesmf_loaded = True
except ImportError:
    xesmf_loaded = False


def test_reset_temporal_axis(PM_da_control_3d_full):
    """Test whether correct new labels are set."""
    smooth = 10
    tsmooth_kws = {"time": smooth}
    first_ori = PM_da_control_3d_full.time[0].values
    first_actual = _reset_temporal_axis(
        PM_da_control_3d_full, tsmooth_kws=tsmooth_kws, dim="time"
    ).time.values[0]
    first_expected = f"{first_ori}-{first_ori+smooth*1-1}"
    assert first_actual == first_expected


def test_reset_temporal_axis_lead(PM_da_initialized_3d_full):
    """Test whether correct new labels are set."""
    smooth = 10
    dim = "lead"
    tsmooth_kws = {dim: smooth}
    first_ori = PM_da_initialized_3d_full.lead[0].values
    first_actual = _reset_temporal_axis(
        PM_da_initialized_3d_full, tsmooth_kws=tsmooth_kws
    )[dim].values[0]
    first_expected = f"{first_ori}-{first_ori+smooth*1-1}"
    assert first_actual == first_expected


def test_temporal_smoothing_reduce_length(PM_da_control_3d_full):
    """Test whether dimsize is reduced properly."""
    smooth = 10
    tsmooth_kws = {"time": smooth}
    actual = temporal_smoothing(
        PM_da_control_3d_full, tsmooth_kws=tsmooth_kws
    ).time.size
    expected = PM_da_control_3d_full.time.size - smooth + 1
    assert actual == expected


@pytest.mark.skipif(not xesmf_loaded, reason="xesmf not installed")
def test_spatial_smoothing_xesmf_reduce_spatial_dims_MPI_curv(
    PM_da_control_3d_full,
):
    """Test whether spatial dimsizes are properly reduced."""
    da = PM_da_control_3d_full
    step = 5
    actual = spatial_smoothing_xesmf(
        da,
        d_lon_lat_kws={"lon": step},
    )
    expected_lat_size = 180 // step
    assert actual["lon"].size < da.lon.size
    assert actual["lat"].size == expected_lat_size


@pytest.mark.skipif(not xesmf_loaded, reason="xesmf not installed")
def test_spatial_smoothing_xesmf_reduce_spatial_dims_CESM(
    reconstruction_ds_3d_full,
):
    """Test whether spatial dimsizes are properly reduced."""
    da = reconstruction_ds_3d_full
    step = 0.1
    actual = spatial_smoothing_xesmf(
        da,
        d_lon_lat_kws={"lat": step},
    )
    # test whether upsampled
    assert actual["lon"].size >= da.nlon.size
    assert actual["lat"].size >= da.nlat.size


def test_smooth_goddard_2013(PM_da_control_3d_full):
    """Test whether Goddard 2013 recommendations are fulfilled by
    smooth_Goddard_2013."""
    da = PM_da_control_3d_full
    actual = smooth_goddard_2013(
        da,
    )
    # test that x, y not in dims
    assert "x" not in actual.dims
    assert "y" not in actual.dims
    # tests whether nlat, nlon got reduced
    assert actual.time.size < da.time.size
    assert actual.lon.size < da.lon.size
    assert actual.lat.size < da.lat.size


def test_compute_after_smooth_goddard_2013(
    PM_da_initialized_3d_full, PM_da_control_3d_full
):
    """Test compute_perfect_model works after smoothings."""
    PM_da_control_3d_full = smooth_goddard_2013(
        PM_da_control_3d_full,
    )
    PM_da_initialized_3d_full = smooth_goddard_2013(
        PM_da_initialized_3d_full,
    )
    actual = compute_perfect_model(PM_da_initialized_3d_full, PM_da_control_3d_full)
    north_atlantic = actual.sel(lat=slice(40, 50), lon=slice(-30, -20))
    assert not north_atlantic.isnull().any()


@pytest.mark.parametrize("smooth", [2, 4])
@pytest.mark.parametrize(
    "pm",
    [
        pytest.lazy_fixture("perfectModelEnsemble_initialized_control_1d_ym_cftime"),
        pytest.lazy_fixture("perfectModelEnsemble_initialized_control_1d_mm_cftime"),
        pytest.lazy_fixture("perfectModelEnsemble_initialized_control_1d_dm_cftime"),
    ],
)
def test_PerfectModelEnsemble_temporal_smoothing_cftime_and_skill(pm, smooth):
    """Test that PredictionEnsemble.smooth({'lead': int}) aggregates lead."""
    pm = pm.isel(lead=range(6))
    pm_smoothed = pm.smooth({"lead": smooth})
    assert (
        pm_smoothed.get_initialized().lead.size
        == pm.get_initialized().lead.size - smooth + 1
    )
    assert pm_smoothed._temporally_smoothed
    skill = pm_smoothed.verify(metric="acc", comparison="m2e", dim=["member", "init"])
    assert skill.lead.size == pm.get_initialized().lead.size - smooth + 1
    assert skill.lead[0] == f"1-{1+smooth-1}"


@pytest.mark.parametrize("dim", ["time", "lead"])
@pytest.mark.parametrize("smooth", [2, 4])
@pytest.mark.parametrize(
    "he",
    [
        pytest.lazy_fixture("hindcast_recon_1d_ym"),
        pytest.lazy_fixture("hindcast_recon_1d_mm"),
        pytest.lazy_fixture("hindcast_recon_1d_dm"),
    ],
)
def test_HindcastEnsemble_temporal_smoothing_cftime_and_skill(he, smooth, dim):
    """Test that HindcastEnsemble.smooth({dim: int}) aggregates lead regardless whether
    time or lead is given as dim."""
    he_smoothed = he.smooth({dim: smooth})
    assert (
        he_smoothed.get_initialized().lead.size
        == he.get_initialized().lead.size - smooth + 1
    )
    skill = he_smoothed.verify(
        metric="acc", comparison="e2o", alignment="maximize", dim="init"
    )
    assert skill.lead.size == he.get_initialized().lead.size - smooth + 1
    assert skill.lead[0] == f"1-{1+smooth-1}"


@pytest.mark.parametrize("step", [1, 2])
@pytest.mark.parametrize("dim", [["lon"], ["lat"], ["lon", "lat"]])
def test_HindcastEnsemble_spatial_smoothing_dim_and_skill(hindcast_recon_3d, dim, step):
    """Test that HindcastEnsemble.smooth({dim: int}) aggregates dim."""
    he = hindcast_recon_3d
    smooth_kws = {key: step for key in dim}
    he_smoothed = he.smooth(smooth_kws)
    for d in dim:
        assert he_smoothed.get_initialized()[d].any()
        assert he_smoothed.get_observations()[d].any()
    assert he_smoothed.verify(
        metric="acc", comparison="e2o", alignment="same_verif", dim="init"
    ).any()


def test_temporal_smoothing_how(perfectModelEnsemble_initialized_control_1d_ym_cftime):
    """Test that PerfectModelEnsemble can smooth by mean and sum aggregation."""
    pm = perfectModelEnsemble_initialized_control_1d_ym_cftime
    pm_smoothed_mean = pm.smooth({"lead": 4}, how="mean")
    pm_smoothed_sum = pm.smooth({"lead": 4}, how="sum")
    assert (
        pm_smoothed_sum.get_initialized().mean()
        > pm_smoothed_mean.get_initialized().mean() * 2
    )


def test_spatial_smoothing_xesmf(hindcast_recon_3d):
    """Test different regridding methods from xesmf.regrid kwargs yield different
    results."""
    he = hindcast_recon_3d
    he_bil = he.smooth("goddard", method="bilinear")
    he_patch = he.smooth("goddard", method="patch")
    assert he_bil.get_initialized().mean() != he_patch.get_initialized().mean()


def test_set_center_coord():
    """Test that center coords are set to the middle of the lead range."""
    da = xr.DataArray(np.arange(2), dims="lead", coords={"lead": ["1-3", "2-4"]})
    actual = _set_center_coord(da).lead_center.values
    expected = [2.0, 3.0]
    assert (actual == expected).all()


def test_PerfectModelEnsemble_smooth_carries_lead_attrs(
    perfectModelEnsemble_initialized_control_1d_ym_cftime,
):
    """Test that PerfectModelEnsemble carries lead attrs after smooth  and verify."""
    pm = perfectModelEnsemble_initialized_control_1d_ym_cftime
    pm_smooth = pm.smooth({"lead": 4}, how="mean")
    assert (
        pm_smooth.verify(metric="rmse", comparison="m2e", dim="init").lead.attrs[
            "units"
        ]
        == "years"
    )


def test_HindcastEnsemble_smooth_carries_lead_attrs(hindcast_recon_1d_ym):
    """Test that HindcastEnsemble carries lead attrs after smooth  and verify."""
    he = hindcast_recon_1d_ym
    he_smooth = he.smooth({"lead": 4}, how="mean")
    assert (
        he_smooth.verify(
            metric="rmse", comparison="e2o", dim="init", alignment="same_verifs"
        ).lead.attrs["units"]
        == "years"
    )


@pytest.mark.parametrize(
    "smooth", [{"lead": 4, "lon": 5, "lat": 5}, "goddard", "goddard2013"]
)
def test_PredictionEnsemble_goddard(
    perfectModelEnsemble_initialized_control_1d_ym_cftime, smooth
):
    """Test that PredictionEnsemble.smooth() understands goodard keys and does multiple
    smoothings in one call."""
    pm = perfectModelEnsemble_initialized_control_1d_ym_cftime
    assert pm.smooth(smooth)


def test_PredictionEnsemble_smooth_None(
    perfectModelEnsemble_initialized_control_1d_ym_cftime,
):
    """Test that PredictionEnsemble.smooth(None) does nothing."""
    pm = perfectModelEnsemble_initialized_control_1d_ym_cftime
    pm_smoothed = pm.smooth(None)
    assert_PredictionEnsemble(pm, pm_smoothed)



import warnings
from functools import wraps

import dask
import xarray as xr

from .constants import (
    CF_LONG_NAMES,
    CF_STANDARD_NAMES,
    CLIMPRED_ENSEMBLE_DIMS,
    VALID_ALIGNMENTS,
    VALID_LEAD_UNITS,
    VALID_REFERENCES,
)
from .exceptions import DatasetError, DimensionError, VariableError
from .options import OPTIONS

NCPU = dask.system.CPU_COUNT


def dec_args_kwargs(wrapper):
    # https://stackoverflow.com/questions/10610824/
    # python-shortcut-for-writing-decorators-which-accept-arguments
    return lambda *dec_args, **dec_kwargs: lambda func: wrapper(
        func, *dec_args, **dec_kwargs
    )


# --------------------------------------#
# CHECKS
# --------------------------------------#
def has_dataset(obj, kind, what):
    """Checks that the PredictionEnsemble has a specific dataset in it."""
    if len(obj) == 0:
        raise DatasetError(
            f"You need to add at least one {kind} dataset before "
            f"attempting to {what}."
        )
    return True


def has_dims(xobj, dims, kind):
    """
    Checks that at the minimum, the object has provided dimensions.
    """
    if isinstance(dims, str):
        dims = [dims]

    if not all(dim in xobj.dims for dim in dims):
        raise DimensionError(
            f"Your {kind} object must contain the "
            f"following dimensions at the minimum: {dims}"
            f", found {list(xobj.dims)}."
        )
    return True


def has_min_len(arr, len_, kind):
    """
    Checks that the array is at least the specified length.
    """
    arr_len = len(arr)
    if arr_len < len_:
        raise DimensionError(
            f"Your {kind} array must be at least {len_}, "
            f"but has only length {arr_len}!"
        )
    return True


def has_valid_lead_units(xobj):
    """
    Checks that the object has valid units for the lead dimension.
    """
    LEAD_UNIT_ERROR = (
        "The lead dimension must must have a valid "
        f"units attribute. Valid options are: {VALID_LEAD_UNITS}"
    )
    # Use `hasattr` here, as it doesn't throw an error if `xobj` doesn't have a
    # coordinate for lead.
    if hasattr(xobj["lead"], "units"):

        units = xobj["lead"].attrs["units"]

        # Check if letter s is appended to lead units string and add it if needed
        if not units.endswith("s"):
            units += "s"
            xobj["lead"].attrs["units"] = units
            warnings.warn(
                f'The letter "s" was appended to the lead units; now {units}.'
            )

        # Raise Error if lead units is not valid
        if not xobj["lead"].attrs["units"] in VALID_LEAD_UNITS:
            raise AttributeError(LEAD_UNIT_ERROR)
    else:
        raise AttributeError(LEAD_UNIT_ERROR)
    return True


def is_in_list(item, list_, kind):
    """Check whether an item is in a list; kind is just a string."""
    if item not in list_:
        raise KeyError(f"Specify {kind} from {list_}: got {item}")
    return True


@dec_args_kwargs
def is_xarray(func, *dec_args):
    """
    Decorate a function to ensure the first arg being submitted is
    either a Dataset or DataArray.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            ds_da_locs = dec_args[0]
            if not isinstance(ds_da_locs, list):
                ds_da_locs = [ds_da_locs]

            for loc in ds_da_locs:
                if isinstance(loc, int):
                    ds_da = args[loc]
                elif isinstance(loc, str):
                    ds_da = kwargs[loc]

                is_ds_da = isinstance(ds_da, (xr.Dataset, xr.DataArray))
                if not is_ds_da:
                    typecheck = type(ds_da)
                    raise IOError(
                        f"""The input data is not an xarray DataArray or
                        Dataset. climpred is built to wrap xarray to make
                        use of its awesome features. Please input an xarray
                        object and retry the function.

                        Your input was of type: {typecheck}"""
                    )
        except IndexError:
            pass
        # this is outside of the try/except so that the traceback is relevant
        # to the actual function call rather than showing a simple Exception
        # (probably IndexError from trying to subselect an empty dec_args list)
        return func(*args, **kwargs)

    return wrapper


def match_calendars(
    ds1, ds2, ds1_time="init", ds2_time="time", kind1="initialized", kind2="observation"
):
    """Checks that calendars match between two xarray Datasets.

    This assumes that the two datasets coming in have cftime time axes.

    Args:
        ds1, ds2 (xarray object): Datasets/DataArrays to compare calendars on. For
            classes, ds1 can be thought of Dataset already existing in the object,
            and ds2 the one being added.
        ds1_time, ds2_time (str, default 'time'): Name of time dimension to look
            for calendar in.
        kind1, kind2 (str): Puts `ds1` and `ds2` into context for custom error message.
            Defaults to `ds1` being "initialized" and `ds2` being "observation".

    Returns:
        True if calendars match, False if they do not match.
    """
    ds1_calendar = ds1[ds1_time].to_index().calendar
    ds2_calendar = ds2[ds2_time].to_index().calendar
    if ds1_calendar != ds2_calendar:
        raise ValueError(
            f"{kind2} calendar of type '{ds2_calendar}' does not match "
            f"{kind1} calendar of type '{ds1_calendar}'. Please modify {kind2} calendar"
            f' using, e.g. `xr.cftime_range(..., calendar="{ds1_calendar}")` and then '
            "try again."
        )


def match_initialized_dims(init, verif, uninitialized=False):
    """Checks that the verification data dimensions match appropriate initialized
    dimensions.

    If uninitialized, ignore ``member``. Otherwise, ignore ``lead`` and ``member``.
    """
    # Since verification data won't have the ``init``` dimension, temporarily rename to
    # time.
    init = init.rename({"init": "time"})
    init_dims = list(init.dims)
    if "lead" in init_dims:
        init_dims.remove("lead")
    if ("member" in init_dims) and not uninitialized:
        init_dims.remove("member")
    if (set(verif.dims) - set(init_dims)) != set():
        unmatch_dims = set(verif.dims) ^ set(init_dims)
        raise DimensionError(
            f"Verification contains more dimensions than initialized. These dimensions do not match: {unmatch_dims}."
        )
    if (set(init_dims) - set(verif.dims)) != set():
        warnings.warn(
            f"Initialized contains more dimensions than verification. Dimension(s) {set(init_dims) - set(verif.dims)} will be broadcasted."
        )
    return True


def match_initialized_vars(init, verif):
    """Checks that a new verification dataset has at least one variable
    in common with the initialized dataset.

    This ensures that they can be compared pairwise.

    Args:
        init (xarray object): Initialized forecast ensemble.
        verif (xarray object): Product to be added.
    """
    init_vars = init.data_vars
    verif_vars = verif.data_vars
    # https://stackoverflow.com/questions/10668282/
    # one-liner-to-check-if-at-least-one-item-in-list-exists-in-another-list
    if set(init_vars).isdisjoint(verif_vars):
        raise VariableError(
            "Please provide a Dataset/DataArray with at least "
            "one matching variable to the initialized prediction ensemble; "
            f"got {init_vars} for init and {verif_vars} for verif."
        )
    return True


def attach_long_names(xobj):
    """Attach long_name for all CLIMPRED_ENSEMBLE_DIMS."""
    xobj2 = xobj.copy()
    for key, value in CF_LONG_NAMES.items():
        if key in xobj2.coords:
            xobj2.coords[key].attrs["long_name"] = value
    return xobj2


def attach_standard_names(xobj):
    """Attach standard_name for all CLIMPRED_ENSEMBLE_DIMS."""
    xobj2 = xobj.copy()
    for key, value in CF_STANDARD_NAMES.items():
        if key in xobj2.coords:
            xobj2.coords[key].attrs["standard_name"] = value
    return xobj2


def rename_to_climpred_dims(xobj):
    """Rename initialized dataset to climpred dims if CF standard_names match."""
    for climpred_d in CF_STANDARD_NAMES.keys():
        cf_standard_name = CF_STANDARD_NAMES[climpred_d]
        if climpred_d not in xobj.dims:
            for d in xobj.dims:
                if xobj[d].attrs.get("standard_name") == cf_standard_name:
                    xobj = xobj.rename({d: climpred_d})
                    xobj[climpred_d].attrs["standard_name"] = cf_standard_name
                    if OPTIONS["warn_for_rename_to_climpred_dims"]:
                        warnings.warn(
                            f'Did not find dimension "{climpred_d}", but renamed dimension {d} with CF-complying standard_name "{cf_standard_name}" to {climpred_d}.'
                        )
    if not set(["init", "lead"]).issubset(set(xobj.dims)):
        warnings.warn(
            f'Could not find dimensions ["init", "lead"] in initialized, found dimension {xobj.dims}. Also searched coordinates for CF-complying standard_names {CF_STANDARD_NAMES}.'
        )
    return xobj


def warn_if_chunking_would_increase_performance(ds, crit_size_in_MB=100):
    """Warn when chunking might make sense.

    Criteria for potential performance increase:
    - input xr.object needs to be chunked realistically.
    - input xr.object needs to sufficiently large so dask overhead doesn't
     overcompensate parallel computation speedup.
    - there should be several CPU available for the computation, like on a
     cluster or multi-core computer
    """
    nbytes_in_MB = ds.nbytes / (1024 ** 2)
    if not dask.is_dask_collection(ds):
        if nbytes_in_MB > crit_size_in_MB and NCPU >= 4:
            warnings.warn(
                "Consider chunking input `ds` along other dimensions than "
                "needed by algorithm, e.g. spatial dimensions, for parallelized "
                "performance increase."
            )
    else:
        if nbytes_in_MB < crit_size_in_MB:
            warnings.warn(
                "Chunking might not bring parallelized performance increase, "
                f"because input size quite small, found {nbytes_in_MB} MB <"
                f" {crit_size_in_MB} MB."
            )
        if NCPU < 4:
            warnings.warn(
                f"Chunking might not bring parallelized performance increase, "
                f"because only few CPUs available, found {NCPU} CPUs."
            )
        number_of_chunks = (
            ds.data.npartitions
            if isinstance(ds, xr.DataArray)
            else ds.to_array().data.npartitions
        )
        if number_of_chunks > 16 * NCPU:
            # much larger than nworkers, warn smaller chunks
            warnings.warn(
                f"Chunking might not bring parallelized performance increase, "
                f"because of much more chunks than CPUs, found {number_of_chunks} "
                f"chunks and {NCPU} CPUs."
            )


def _check_valid_reference(reference):
    """Enforce reference as list and check for valid entries."""
    if reference is None:
        reference = []
    if isinstance(reference, str):
        reference = [reference]
    if not isinstance(reference, list):
        reference = list(reference)
    if not set(reference).issubset(set(VALID_REFERENCES)):
        raise ValueError(
            f"Specify reference from {VALID_REFERENCES}, found {reference}"
        )
    return reference


def _check_valud_alignment(alignment):
    if alignment not in VALID_ALIGNMENTS:
        raise ValueError(
            f"Please provide alignment from {VALID_ALIGNMENTS}, found alignment='{alignment}'."
        )
    else:
        if alignment == "same_init":
            alignment = "same_inits"
        elif alignment == "same_verifs":
            alignment = "same_verif"
    return alignment



# Configuration file for the Sphinx documentation builder.
#
# u This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

import datetime
import os
import sys

import climpred

sys.path.insert(0, os.path.abspath("../.."))


# -- Project information -----------------------------------------------------
current_year = datetime.datetime.now().year
project = "climpred"
copyright = f"2019-{current_year}, climpred development team"
author = "climpred development team"

version = climpred.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.napoleon",
    "sphinx.ext.imgmath",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

# Cupybutton configuration
# See: https://sphinx-copybutton.readthedocs.io/en/latest/
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_only_copy_prompt_lines = True
copybutton_remove_prompts = True

extlinks = {
    "issue": ("https://github.com/pangeo-data/climpred/issues/%s", "GH#"),
    "pr": ("https://github.com/pangeo-data/climpred/pull/%s", "GH#"),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["build", "**.ipynb_checkpoints", "Thumbs.db", ".DS_Store"]

pygments_style = "sphinx"
source_suffix = ".rst"
master_doc = "index"

nbsphinx_timeout = 180  # 3 minute timeout
# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = "alabaster"
html_theme = "sphinx_rtd_theme"
html_logo = "images/climpred-logo.png"
html_theme_options = {"logo_only": False, "style_nav_header_background": "#fcfcfc"}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "xarray": ("https://xarray.pydata.org/en/stable/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "xskillscore": ("https://xskillscore.readthedocs.io/en/stable", None),
}

# Should only be uncommented when testing page development while notebooks
# are breaking.
# nbsphinx_allow_errors = True



import numpy as np
import pytest
import xarray as xr

from climpred import PerfectModelEnsemble
from climpred.exceptions import DatasetError
from climpred.metrics import DETERMINISTIC_PM_METRICS

xr.set_options(display_style="text")

comparison_dim_PM = [
    ("m2m", "init"),
    ("m2m", "member"),
    ("m2m", ["init", "member"]),
    ("m2e", "init"),
    ("m2e", "member"),
    ("m2e", ["init", "member"]),
    ("m2c", "init"),
    ("m2c", "member"),
    ("m2c", ["init", "member"]),
    ("e2c", "init"),
]

references = [
    "uninitialized",
    "persistence",
    "climatology",
    ["climatology", "uninitialized", "persistence"],
]
references_ids = [
    "uninitialized",
    "persistence",
    "climatology",
    "climatology, uninitialized, persistence",
]

category_edges = np.array([9.5, 10.0, 10.5])

ITERATIONS = 3


def test_perfectModelEnsemble_init(PM_ds_initialized_1d):
    """Test to see if perfect model ensemble can be initialized"""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d)
    print(PerfectModelEnsemble)
    assert pm


def test_perfectModelEnsemble_init_da(PM_da_initialized_1d):
    """Test to see if perfect model ensemble can be initialized with da"""
    pm = PerfectModelEnsemble(PM_da_initialized_1d)
    assert pm


def test_add_control(perfectModelEnsemble_initialized_control):
    """Test to see if control can be added to PerfectModelEnsemble"""
    assert perfectModelEnsemble_initialized_control.get_control()


def test_generate_uninit(perfectModelEnsemble_initialized_control):
    """Test to see if uninitialized ensemble can be bootstrapped"""
    pm = perfectModelEnsemble_initialized_control
    pm = pm.generate_uninitialized()
    assert pm.get_uninitialized()


def test_compute_persistence(perfectModelEnsemble_initialized_control):
    """Test that compute persistence can be run for perfect model ensemble"""
    perfectModelEnsemble_initialized_control._compute_persistence(metric="acc")


def test_get_initialized(PM_ds_initialized_1d):
    """Test whether get_initialized function works."""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d)
    init = pm.get_initialized()
    assert init == pm._datasets["initialized"]


def test_get_uninitialized(perfectModelEnsemble_initialized_control):
    """Test whether get_uninitialized function works."""
    pm = perfectModelEnsemble_initialized_control
    pm = pm.generate_uninitialized()
    uninit = pm.get_uninitialized()
    assert uninit == pm._datasets["uninitialized"]


def test_get_control(perfectModelEnsemble_initialized_control):
    """Test whether get_control function works."""
    ctrl = perfectModelEnsemble_initialized_control.get_control()
    assert ctrl == perfectModelEnsemble_initialized_control._datasets["control"]


def test_inplace(PM_ds_initialized_1d, PM_ds_control_1d):
    """Tests that inplace operations do not work."""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d)
    # Adding a control.
    pm.add_control(PM_ds_control_1d)
    with_ctrl = pm.add_control(PM_ds_control_1d)
    assert pm != with_ctrl
    # Adding an uninitialized ensemble.
    pm = pm.add_control(PM_ds_control_1d)
    pm.generate_uninitialized()
    with_uninit = pm.generate_uninitialized()
    assert pm != with_uninit
    # Applying arbitrary func.
    pm.sum("init")
    summed = pm.sum("init")
    assert pm != summed


def test_verify(perfectModelEnsemble_initialized_control):
    """Test that verify works."""
    assert perfectModelEnsemble_initialized_control.verify(
        metric="mse", comparison="m2e", dim=["init", "member"]
    )


def test_verify_metric_kwargs(perfectModelEnsemble_initialized_control):
    """Test that verify with metric_kwargs works."""
    pm = perfectModelEnsemble_initialized_control
    pm = pm - pm.mean("time").mean("init")
    assert pm.verify(
        metric="threshold_brier_score",
        comparison="m2c",
        dim=["init", "member"],
        threshold=0.5,
    )


@pytest.mark.parametrize("reference", references, ids=references_ids)
def test_verify_reference(perfectModelEnsemble_initialized_control, reference):
    """Test that verify works with references given."""
    pm = perfectModelEnsemble_initialized_control.generate_uninitialized()
    skill = (
        pm.verify(
            metric="rmse", comparison="m2e", dim=["init", "member"], reference=reference
        )
        .expand_dims(["lon", "lat"])
        .isel(lon=[0] * 2, lat=[0] * 2)
    )  # make geospatial
    if isinstance(reference, str):
        reference = [reference]
    elif reference is None:
        reference = []
    if len(reference) == 0:
        assert "skill" not in skill.dims
    else:
        assert skill.skill.size == len(reference) + 1
    # test skills not none
    assert skill.notnull().all()
    assert "dayofyear" not in skill.coords


def test_verify_fails_expected_metric_kwargs(perfectModelEnsemble_initialized_control):
    """Test that verify without metric_kwargs fails."""
    pm = perfectModelEnsemble_initialized_control
    pm = pm - pm.mean("time").mean("init")
    with pytest.raises(ValueError) as excinfo:
        pm.verify(
            metric="threshold_brier_score", comparison="m2c", dim=["init", "member"]
        )
    assert "Please provide threshold." == str(excinfo.value)


def test_compute_uninitialized_metric_kwargs(perfectModelEnsemble_initialized_control):
    "Test that _compute_uninitialized with metric_kwargs works"
    pm = perfectModelEnsemble_initialized_control
    pm = pm - pm.mean("time").mean("init")
    pm = pm.generate_uninitialized()
    assert pm._compute_uninitialized(
        metric="threshold_brier_score",
        comparison="m2c",
        threshold=0.5,
        dim=["init", "member"],
    )


def test_bootstrap_metric_kwargs(perfectModelEnsemble_initialized_control):
    """Test that bootstrap with metric_kwargs works."""
    pm = perfectModelEnsemble_initialized_control
    pm = pm - pm.mean("time").mean("init")
    pm = pm.generate_uninitialized()
    assert pm.bootstrap(
        metric="threshold_brier_score",
        comparison="m2c",
        threshold=0.5,
        iterations=ITERATIONS,
        dim=["init", "member"],
    )


def test_calendar_matching_control(PM_da_initialized_1d, PM_ds_control_1d):
    """Tests that error is thrown if calendars mismatch when adding observations."""
    pm = PerfectModelEnsemble(PM_da_initialized_1d)
    PM_ds_control_1d["time"] = xr.cftime_range(
        start="1950", periods=PM_ds_control_1d.time.size, freq="MS", calendar="all_leap"
    )
    with pytest.raises(ValueError) as excinfo:
        pm = pm.add_control(PM_ds_control_1d)
    assert "does not match" in str(excinfo.value)


def test_persistence_dim(perfectModelEnsemble_initialized_control):
    pm = perfectModelEnsemble_initialized_control.expand_dims(
        "lon"
    ).generate_uninitialized()
    assert "lon" in pm.get_initialized().dims
    dim = ["lon"]
    metric = "rmse"
    comparison = "m2e"

    actual = pm._compute_persistence(metric=metric, dim=dim)
    assert "lon" not in actual.dims
    assert "init" in actual.dims

    actual = pm.verify(
        metric=metric,
        comparison=comparison,
        dim=dim,
        reference=["persistence", "uninitialized"],
    )
    assert "lon" not in actual.dims
    assert "init" in actual.dims

    pm = perfectModelEnsemble_initialized_control.expand_dims("lon")
    # fix _resample_iterations_idx doesnt work with singular dimension somewhere.
    pm = pm.isel(lon=[0, 0])
    actual = pm.bootstrap(metric=metric, comparison=comparison, dim=dim, iterations=2)
    assert "lon" not in actual.dims
    assert "init" in actual.dims


def test_HindcastEnsemble_as_PerfectModelEnsemble(hindcast_recon_1d_mm):
    """Test that initialized dataset for HindcastEnsemble can also be used for
    PerfectModelEnsemble."""
    v = "SST"
    alignment = "maximize"
    hindcast = hindcast_recon_1d_mm.isel(lead=[0, 1])
    assert (
        not hindcast.verify(
            metric="acc", comparison="e2o", dim="init", alignment=alignment
        )[v]
        .isnull()
        .any()
    )

    # try PerfectModelEnsemble predictability
    init = hindcast.get_initialized()
    pm = PerfectModelEnsemble(init)

    assert (
        not pm.verify(metric="acc", comparison="m2e", dim=["member", "init"])[v]
        .isnull()
        .any()
    )


def test_verify_no_need_for_control(PM_da_initialized_1d, PM_da_control_1d):
    """Tests that no error is thrown when no control present
    when calling verify(reference=['uninitialized'])."""
    v = "tos"
    comparison = "m2e"
    pm = PerfectModelEnsemble(PM_da_initialized_1d).isel(lead=[0, 1, 2])
    # verify needs to control
    skill = pm.verify(metric="mse", comparison=comparison, dim="init")
    assert not skill[v].isnull().any()
    # control not needed for normalized metrics as normalized
    # with verif which is the verification member in PM and
    # not the control simulation.
    assert (
        not pm.verify(metric="nmse", comparison=comparison, dim="init")[v]
        .isnull()
        .any()
    )

    with pytest.raises(DatasetError) as e:
        pm.verify(
            metric="mse", comparison=comparison, dim="init", reference=["persistence"]
        )
    assert "at least one control dataset" in str(e.value)

    # unlikely case that control gets deleted after generating uninitialized
    pm = pm.add_control(PM_da_control_1d).generate_uninitialized()
    pm._datasets["control"] = {}
    assert (
        not pm._compute_uninitialized(metric="mse", comparison=comparison, dim="init")[
            v
        ]
        .isnull()
        .any()
    )

    assert (
        not pm.verify(
            metric="mse", comparison=comparison, dim="init", reference=["uninitialized"]
        )[v]
        .isnull()
        .any()
    )


def test_verify_reference_same_dims(perfectModelEnsemble_initialized_control):
    """Test that verify returns the same dimensionality regardless of reference."""
    pm = perfectModelEnsemble_initialized_control.generate_uninitialized()
    pm = pm.isel(lead=[0, 1, 2], init=[0, 1, 2])
    metric = "mse"
    comparison = "m2e"
    dim = "init"
    actual_no_ref = pm.verify(
        metric=metric, comparison=comparison, dim=dim, reference=None
    )
    actual_uninit_ref = pm.verify(
        metric=metric, comparison=comparison, dim=dim, reference="uninitialized"
    )
    actual_pers_ref = pm.verify(
        metric=metric, comparison=comparison, dim=dim, reference="persistence"
    )
    actual_clim_ref = pm.verify(
        metric=metric, comparison=comparison, dim=dim, reference="climatology"
    )
    assert actual_uninit_ref.skill.size == 2
    assert actual_pers_ref.skill.size == 2
    assert actual_clim_ref.skill.size == 2
    # no additional dimension, +1 because initialized squeezed
    assert len(actual_no_ref.dims) + 1 == len(actual_pers_ref.dims)
    assert len(actual_no_ref.dims) + 1 == len(actual_uninit_ref.dims)
    assert len(actual_no_ref.dims) + 1 == len(actual_clim_ref.dims)
    assert len(actual_pers_ref.dims) == len(actual_uninit_ref.dims)
    assert len(actual_clim_ref.dims) == len(actual_uninit_ref.dims)


@pytest.mark.parametrize("reference", references, ids=references_ids)
@pytest.mark.parametrize("comparison,dim", comparison_dim_PM)
@pytest.mark.parametrize("metric", DETERMINISTIC_PM_METRICS)
def test_PerfectModel_verify_bootstrap_deterministic(
    perfectModelEnsemble_initialized_control, comparison, metric, dim, reference
):
    """
    Checks that PerfectModel.verify() and PerfectModel.bootstrap() for
    deterministic metrics is not NaN.
    """
    pm = perfectModelEnsemble_initialized_control.isel(lead=[0, 1, 2], init=range(6))
    if isinstance(reference, str):
        reference = [reference]
    if metric == "contingency":
        metric_kwargs = {
            "forecast_category_edges": category_edges,
            "observation_category_edges": category_edges,
            "score": "accuracy",
        }
    elif metric == "roc":
        metric_kwargs = {"bin_edges": category_edges}
    else:
        metric_kwargs = {}
    # acc on dim member only is ill defined
    pearson_r_containing_metrics = [
        "pearson_r",
        "spearman_r",
        "pearson_r_p_value",
        "spearman_r_p_value",
        "msess_murphy",
        "bias_slope",
        "conditional_bias",
        "std_ratio",
        "conditional_bias",
        "uacc",
    ]
    if dim == "member" and metric in pearson_r_containing_metrics:
        dim = ["init", "member"]

    actual = pm.verify(
        comparison=comparison,
        metric=metric,
        dim=dim,
        reference=reference,
        **metric_kwargs
    ).tos
    if metric in ["contingency"] or metric in pearson_r_containing_metrics:
        # less strict here with all NaNs, pearson_r yields NaNs for climatology
        if "climatology" in reference:
            actual = actual.drop_sel(skill="climatology")
        assert not actual.isnull().all()
    else:
        assert not actual.isnull().any()

    # bootstrap()
    actual = pm.bootstrap(
        comparison=comparison,
        metric=metric,
        dim=dim,
        iterations=ITERATIONS,
        reference=reference,
        **metric_kwargs
    ).tos
    if len(reference) > 0:
        actual = actual.drop_sel(results="p")

    if metric in ["contingency"] or metric in pearson_r_containing_metrics:
        # less strict here with all NaNs, pearson_r yields NaNs for climatology
        if "climatology" in reference:
            actual = actual.drop_sel(skill="climatology")
        assert not actual.sel(results="verify skill").isnull().all()
    else:
        assert not actual.sel(results="verify skill").isnull().any()


@pytest.mark.parametrize("metric", ("rmse", "pearson_r"))
def test_pvalue_from_bootstrapping(perfectModelEnsemble_initialized_control, metric):
    """Test that pvalue of initialized ensemble first lead is close to 0."""
    sig = 95
    pm = perfectModelEnsemble_initialized_control.isel(lead=[0, 1, 2])
    actual = (
        pm.bootstrap(
            metric=metric,
            iterations=ITERATIONS,
            comparison="e2c",
            sig=sig,
            dim="init",
            reference="uninitialized",
        )
        .sel(skill="uninitialized", results="p")
        .isel(lead=0)
    )
    # check that significant p-value
    assert actual.tos.values < 2 * (1 - sig / 100)
    # lead units keep
    assert actual.lead.attrs["units"] == "years"



import datetime
import logging
import warnings

import cftime
import dask
import numpy as np
import pandas as pd
import xarray as xr
from xarray.coding.cftime_offsets import to_offset

from . import comparisons, metrics
from .checks import is_in_list
from .comparisons import COMPARISON_ALIASES
from .constants import FREQ_LIST_TO_INFER_STRIDE, HINDCAST_CALENDAR_STR
from .metrics import METRIC_ALIASES
from .options import OPTIONS


def assign_attrs(
    skill,
    ds,
    function_name=None,
    metadata_dict=None,
    alignment=None,
    metric=None,
    comparison=None,
    dim=None,
):
    """Write information about prediction skill into attrs.

    Args:
        skill (`xarray` object): prediction skill.
        ds (`xarray` object): prediction ensemble with inits.
        function_name (str): name of compute function
        metadata_dict (dict): optional attrs
        alignment (str): method used to align inits and verification data.
        metric (class) : metric used in comparing the forecast and verification data.
        comparison (class): how to compare the forecast and verification data.
        dim (str): Dimension over which metric was applied.

    Returns:
       skill (`xarray` object): prediction skill with additional attrs.
    """
    # assign old attrs
    skill.attrs = ds.attrs

    # climpred info
    skill.attrs[
        "prediction_skill"
    ] = "calculated by climpred https://climpred.readthedocs.io/"
    if function_name:
        skill.attrs["skill_calculated_by_function"] = function_name
    if "init" in ds.coords:
        skill.attrs["number_of_initializations"] = ds.init.size
    if "member" in ds.coords and function_name != "compute_persistence":
        skill.attrs["number_of_members"] = ds.member.size

    if alignment is not None:
        skill.attrs["alignment"] = alignment
    skill.attrs["metric"] = metric.name
    if comparison is not None:
        skill.attrs["comparison"] = comparison.name
    if dim is not None:
        skill.attrs["dim"] = dim

    # change unit power
    if metric.unit_power == 0:
        skill.attrs["units"] = "None"
    if metric.unit_power >= 2 and "units" in skill.attrs:
        p = metric.unit_power
        p = int(p) if int(p) == p else p
        skill.attrs["units"] = f"({skill.attrs['units']})^{p}"

    # check for none attrs and remove
    del_list = []
    for key, value in metadata_dict.items():
        if value is None and key != "units":
            del_list.append(key)
    for entry in del_list:
        del metadata_dict[entry]

    # write optional information
    if metadata_dict is None:
        metadata_dict = dict()
    skill.attrs.update(metadata_dict)

    return skill


def convert_time_index(xobj, time_string, kind, calendar=HINDCAST_CALENDAR_STR):
    """Converts incoming time index to a standard xr.CFTimeIndex.

    Args:
        xobj (xarray object): Dataset or DataArray with a time dimension to convert.
        time_string (str): Name of time dimension.
        kind (str): Kind of object for error message.
        calendar (str): calendar to set time dimension to.

    Returns:
        Dataset or DataArray with converted time dimension. If incoming time index is
        ``xr.CFTimeIndex``, returns the same index. If ``pd.DatetimeIndex``, converts to
        ``cftime.ProlepticGregorian``. If ``pd.Int64Index`` or ``pd.Float64Index``,
        assumes annual resolution and returns year-start ``cftime.ProlepticGregorian``.

    Raises:
        ValueError: If ``time_index`` is not an ``xr.CFTimeIndex``, ``pd.Int64Index``,
            ``pd.Float64Index``, or ``pd.DatetimeIndex``.
    """
    xobj = xobj.copy()  # Ensures that main object index is not overwritten.
    time_index = xobj[time_string].to_index()

    if not isinstance(time_index, xr.CFTimeIndex):

        if isinstance(time_index, pd.DatetimeIndex):
            # Extract year, month, day strings from datetime.
            time_strings = [str(t) for t in time_index]
            split_dates = [d.split(" ")[0].split("-") for d in time_strings]
            split_dates = [
                d.split("-") for d in time_index.strftime("%Y-%m-%d-%H-%M-%S")
            ]

        # If Float64Index or Int64Index, assume annual and convert accordingly.
        elif isinstance(time_index, pd.Float64Index) | isinstance(
            time_index, pd.Int64Index
        ):
            if OPTIONS["warn_for_init_coords_int_to_annual"]:
                warnings.warn(
                    "Assuming annual resolution due to numeric inits. "
                    "Change init to a datetime if it is another resolution."
                )
            # TODO: What about decimal time? E.g. someone has 1955.5 or something?
            # hard to maintain a clear rule below seasonality
            dates = [str(int(t)) + "-01-01-00-00-00" for t in time_index]
            split_dates = [d.split("-") for d in dates]
            if "lead" in xobj.dims:
                # Probably the only case we can assume lead units, since `lead` does not
                # give us any information on this.
                xobj["lead"].attrs["units"] = "years"

        else:
            raise ValueError(
                f"Your {kind} object must be pd.Float64Index, "
                "pd.Int64Index, xr.CFTimeIndex or "
                "pd.DatetimeIndex."
            )
        cftime_dates = [
            getattr(cftime, calendar)(int(y), int(m), int(d), int(H), int(M), int(S))
            for (y, m, d, H, M, S) in split_dates
        ]
        time_index = xr.CFTimeIndex(cftime_dates)
        xobj[time_string] = time_index

    return xobj


def convert_cftime_to_datetime_coords(ds, dim):
    """Convert dimension coordinate dim from CFTimeIndex to pd.DatetimeIndex."""
    return ds.assign_coords(
        {dim: xr.DataArray(ds[dim].to_index().to_datetimeindex(), dims=dim)}
    )


def find_start_dates_for_given_init(control, single_init):
    """Find the same start dates for cftime single_init across different years in
    control. Return control.time. Requires calendar=Datetime(No)Leap for consistent
    `dayofyear`."""
    # check that Leap or NoLeap calendar
    for dim in [single_init.init, control.time]:
        # dirty workaround .values requires a dimension but single_init is only a
        # single initialization and therefore without init dim
        dim = dim.expand_dims("init") if "time" not in dim.coords else dim
        calendar = type(dim.values[0]).__name__
        if "Leap" not in calendar:
            warnings.warn(
                f"inputs to `find_start_dates_for_given_init` should be `Leap` "
                f" or `NoLeap` calendar, found {calendar} in {dim}."
                f" Otherwise dayofyear is not static and can lead to slight shifts."
            )
    # could also just take first of month or even a random number day in month
    take_same_time = "dayofyear"
    return control.sel(
        time=getattr(control.time.dt, take_same_time).values
        == getattr(single_init.init.dt, take_same_time).values
    ).time


def return_time_series_freq(ds, dim):
    """Return the temporal frequency of the input time series. Finds the frequency
    starting from high frequencies at which all ds.dim are not equal.

    Args:
        ds (xr.object): input with dimension `dim`.
        dim (str): name of dimension.

    Returns:
        str: frequency string from FREQ_LIST_TO_INFER_STRIDE

    """
    for freq in FREQ_LIST_TO_INFER_STRIDE:
        # first dim values not equal all others
        if not (
            getattr(ds.isel({dim: 0})[dim].dt, freq) == getattr(ds[dim].dt, freq)
        ).all():
            return freq


def get_metric_class(metric, list_):
    """
    This allows the user to submit a string representing the desired metric
    to the corresponding metric class.

    Currently compatable with functions:
    * compute_persistence()
    * compute_perfect_model()
    * compute_hindcast()

    Args:
        metric (str): name of metric.
        list_ (list): check whether metric in list

    Returns:
        metric (Metric): class object of the metric.
    """
    if isinstance(metric, metrics.Metric):
        return metric
    elif isinstance(metric, str):
        # check if metric allowed
        is_in_list(metric, list_, "metric")
        metric = METRIC_ALIASES.get(metric, metric)
        return getattr(metrics, "__" + metric)
    else:
        raise ValueError(
            f"Please provide metric as str or Metric class, found {type(metric)}"
        )


def get_comparison_class(comparison, list_):
    """
    Converts a string comparison entry from the user into a Comparison class
     for the package to interpret.

    Perfect Model:

        * m2m: Compare all members to all other members.
        * m2c: Compare all members to the control.
        * m2e: Compare all members to the ensemble mean.
        * e2c: Compare the ensemble mean to the control.

    Hindcast:

        * e2o: Compare the ensemble mean to the verification data.
        * m2o: Compare each ensemble member to the verification data.

    Args:
        comparison (str): name of comparison.

    Returns:
        comparison (Comparison): comparison class.

    """
    if isinstance(comparison, comparisons.Comparison):
        return comparison
    elif isinstance(comparison, str):
        # check if comparison allowed
        is_in_list(comparison, list_, "comparison")
        comparison = COMPARISON_ALIASES.get(comparison, comparison)
        return getattr(comparisons, "__" + comparison)
    else:
        is_in_list(comparison, list_, "comparison")
        return getattr(comparisons, "__" + comparison)


def get_lead_cftime_shift_args(units, lead):
    """Determines the date increment to use when adding the lead time to init time based
    on the units attribute.

    Args:
        units (str): Units associated with the lead dimension. Must be
            years, seasons, months, weeks, pentads, days, hours, minutes.
        lead (int): Increment of lead being computed.

    Returns:
       n (int): Number of units to shift. ``value`` for
           ``CFTimeIndex.shift(value, str)``.
       freq (str): Pandas frequency alias. ``str`` for
           ``CFTimeIndex.shift(value, str)``.
    """
    lead = int(lead)

    d = {
        # Currently assumes yearly aligns with year start.
        "years": (lead, "YS"),
        "seasons": (lead * 3, "MS"),
        # Currently assumes monthly aligns with month start.
        "months": (lead, "MS"),
        "weeks": (lead * 7, "D"),
        "pentads": (lead * 5, "D"),
        "days": (lead, "D"),
        "hours": (lead, "H"),
        "minutes": (lead, "T"),
        "seconds": (lead, "S"),
    }

    try:
        n, freq = d[units]
    except KeyError:
        print(f"{units} is not a valid choice.")
        print(f"Accepted `units` values include: {d.keys()}")
    return n, freq


def get_multiple_lead_cftime_shift_args(units, leads):
    """Returns ``CFTimeIndex.shift()`` offset increment for an arbitrary number of
    leads.

    Args:
        units (str): Units associated with the lead dimension. Must be one of
            years, seasons, months, weeks, pentads, days, hours, minutes.
        leads (list, array, xr.DataArray of ints): Leads to return offset for.

    Returns:
        n (tuple of ints): Numbers of units to shift for ``leads``. ``value`` for
            ``CFTimeIndex.shift(value, str)``.
        freq (str): Pandas frequency alias. ``str`` for
            ``CFTimeIndex.shift(value, str)``.
    """
    n_freq_tuples = [get_lead_cftime_shift_args(units, lead) for lead in leads]
    n, freq = list(zip(*n_freq_tuples))
    return n, freq[0]


def intersect(lst1, lst2):
    """
    Custom intersection, since `set.intersection()` changes type of list.
    """
    lst3 = [value for value in lst1 if value in lst2]
    return np.array(lst3)


def lead_units_equal_control_time_stride(init, verif):
    """Check that the lead units of the initialized ensemble have the same frequency as
    the control stride.

    Args:
        init (xr.object): initialized ensemble with lead units.
        verif (xr.object): control, uninitialized historical simulation / observations.

    Returns:
        bool: Possible to continue or raise warning.

    """
    verif_time_stride = return_time_series_freq(verif, "time")
    lead_units = init.lead.attrs["units"].strip("s")
    if verif_time_stride != lead_units:
        raise ValueError(
            "Please provide the same temporal resolution for verif.time",
            f"(found {verif_time_stride}) and init.init (found",
            f"{lead_units}).",
        )
    else:
        return True


def _load_into_memory(res):
    """Compute if res is lazy data."""
    if dask.is_dask_collection(res):
        res = res.compute()
    return res


def rechunk_to_single_chunk_if_more_than_one_chunk_along_dim(ds, dim):
    """Rechunk an xarray object more than one chunk along dim."""
    if dask.is_dask_collection(ds) and dim in ds.chunks:
        if isinstance(ds, xr.Dataset):
            nchunks = len(ds.chunks[dim])
        elif isinstance(ds, xr.DataArray):
            nchunks = len(ds.chunks[ds.get_axis_num(dim)])
        if nchunks > 1:
            ds = ds.chunk({dim: -1})
    return ds


def shift_cftime_index(xobj, time_string, n, freq):
    """Shifts a ``CFTimeIndex`` over a specified number of time steps at a given
    temporal frequency.

    This leverages the handy ``.shift()`` method from ``xarray.CFTimeIndex``. It's a
    simple call, but is used throughout ``climpred`` so it is documented here clearly
    for convenience.

    Args:
        xobj (xarray object): Dataset or DataArray with the ``CFTimeIndex`` to shift.
        time_string (str): Name of time dimension to be shifted.
        n (int): Number of units to shift.
            Returned from :py:func:`get_lead_cftime_shift_args`.
        freq (str): Pandas frequency alias.
            Returned from :py:func:`get_lead_cftime_shift_args`.

    Returns:
        ``CFTimeIndex`` shifted by ``n`` steps at time frequency ``freq``.
    """
    time_index = xobj[time_string].to_index()
    return time_index.shift(n, freq)


def shift_cftime_singular(cftime, n, freq):
    """Shifts a singular ``cftime`` by the desired frequency.

    This directly pulls the ``shift`` method from ``CFTimeIndex`` in ``xarray``. This
    is useful if you need to shift a singular ``cftime`` by some offset, but are not
    working with a full ``CFTimeIndex``.

    Args:
        cftime (``cftime``): ``cftime`` object to shift.
        n (int): Number of steps to shift by.
        freq (str): Frequency string, per ``pandas`` convention.

    See:
    https://github.com/pydata/xarray/blob/master/xarray/coding/cftimeindex.py#L376.
    """
    if not isinstance(n, int):
        raise TypeError(f"'n' must be an int, got {n}.")
    if not isinstance(freq, str):
        raise TypeError(f"'freq' must be a str, got {freq}.")
    return cftime + n * to_offset(freq)


def _transpose_and_rechunk_to(new_chunk_ds, ori_chunk_ds):
    """Chunk xr.object `new_chunk_ds` as another xr.object `ori_chunk_ds`.
    This is needed after some operations which reduce chunks to size 1.
    First transpose a to ds.dims then apply ds chunking to a."""
    transpose_kwargs = (
        {"transpose_coords": False} if isinstance(new_chunk_ds, xr.DataArray) else {}
    )
    return new_chunk_ds.transpose(*ori_chunk_ds.dims, **transpose_kwargs).chunk(
        ori_chunk_ds.chunks
    )


def convert_Timedelta_to_lead_units(ds):
    """Convert lead as pd.Timedelta to lead as int and corresponding lead.attrs['units'] and convert to longest integer lead unit possible."""
    if ds["lead"].dtype == "<m8[ns]":
        ds["lead"] = (ds.lead * 1e-9).astype(int)
        ds["lead"].attrs["units"] = "seconds"

    if (ds["lead"] % 60 == 0).all() and ds["lead"].attrs["units"] == "seconds":
        ds["lead"] = ds["lead"] / 60
        ds["lead"].attrs["units"] = "minutes"
    if (ds["lead"] % 60 == 0).all() and ds["lead"].attrs["units"] == "minutes":
        ds["lead"] = ds["lead"] / 60
        ds["lead"].attrs["units"] = "hours"
    if (ds["lead"] % 24 == 0).all() and ds["lead"].attrs["units"] == "hours":
        ds["lead"] = ds["lead"] / 24
        ds["lead"].attrs["units"] = "days"
    if (ds["lead"] % 5 == 0).all() and ds["lead"].attrs["units"] == "days":
        ds["lead"] = ds["lead"] / 5
        ds["lead"].attrs["units"] = "pentads"
    if (ds["lead"] % 7 == 0).all() and ds["lead"].attrs["units"] == "days":
        ds["lead"] = ds["lead"] / 7
        ds["lead"].attrs["units"] = "weeks"
    return ds


def broadcast_time_grouped_to_time(forecast, category_edges, dim):
    """Broadcast time.groupby('time.month/dayofyear/weekofyear').mean() from
    category_edges back to dim matching forecast."""
    category_edges_time_dim = [
        d
        for d in category_edges.dims
        if d in ["season", "month", "weekofyear", "dayofyear"]
    ]
    if isinstance(category_edges_time_dim, list):
        if len(category_edges_time_dim) > 0:
            category_edges_time_dim = category_edges_time_dim[0]
            logging.debug(f"found category_edges_time_dim = {category_edges_time_dim}")
            category_edges = category_edges.sel(
                {
                    category_edges_time_dim: getattr(
                        forecast[dim].dt, category_edges_time_dim
                    )
                }
            )
    return category_edges


def broadcast_metric_kwargs_for_rps(forecast, verif, metric_kwargs):
    """Apply broadcast_time_grouped_to_time to category_edges in metric_kwargs."""
    category_edges = metric_kwargs.get("category_edges", None)
    logging.debug("enter climpred.utils.broadcast_metric_kwargs_for_rps")
    if category_edges is not None:
        if isinstance(category_edges, tuple):
            logging.debug("category_edges is tuple")
            verif_edges = category_edges[0]
            verif_edges = broadcast_time_grouped_to_time(verif, verif_edges, dim="time")
            forecast_edges = category_edges[1]
            forecast_edges = broadcast_time_grouped_to_time(
                forecast, forecast_edges, dim="init"
            )
            metric_kwargs["category_edges"] = (verif_edges, forecast_edges)
        elif isinstance(category_edges, xr.Dataset):
            logging.debug("category_edges is xr.Dataset")
            metric_kwargs["category_edges"] = broadcast_time_grouped_to_time(
                verif, category_edges, dim="time"
            )
        elif isinstance(category_edges, np.ndarray):
            logging.debug("category_edges is np.array")
        else:
            raise ValueError(
                f"excepted category edges as tuple, xr.Dataset or np.array, found {type(category_edges)}"
            )
        return metric_kwargs



import numpy as np
import xarray as xr
from dask.distributed import Client

from climpred.bootstrap import bootstrap_hindcast
from climpred.metrics import PROBABILISTIC_METRICS
from climpred.prediction import compute_hindcast

from . import ensure_loaded, parameterized, randn, requires_dask

# only take subselection of all possible metrics
METRICS = ["rmse", "pearson_r", "crpss"]
# only take comparisons compatible with probabilistic metrics
HINDCAST_COMPARISONS = ["m2o"]

ITERATIONS = 16


class Generate:
    """
    Generate input data for benchmark.
    """

    timeout = 600
    repeat = (2, 5, 20)

    def make_hind_obs(self):
        """Generates initialized hindcast, uninitialized historical and observational
        data, mimicking a hindcast experiment."""
        self.hind = xr.Dataset()
        self.observations = xr.Dataset()
        self.uninit = xr.Dataset()

        self.nmember = 3
        self.nlead = 5
        self.nx = 72
        self.ny = 36
        self.iterations = ITERATIONS
        self.init_start = 1960
        self.init_end = 2000
        self.ninit = self.init_end - self.init_start
        self.client = None

        FRAC_NAN = 0.0

        inits = xr.cftime_range(
            start=str(self.init_start), end=str(self.init_end - 1), freq="YS"
        )
        leads = np.arange(1, 1 + self.nlead)
        members = np.arange(1, 1 + self.nmember)

        lons = xr.DataArray(
            np.linspace(0.5, 359.5, self.nx),
            dims=("lon",),
            attrs={"units": "degrees east", "long_name": "longitude"},
        )
        lats = xr.DataArray(
            np.linspace(-89.5, 89.5, self.ny),
            dims=("lat",),
            attrs={"units": "degrees north", "long_name": "latitude"},
        )
        self.hind["var"] = xr.DataArray(
            randn(
                (self.nmember, self.ninit, self.nlead, self.nx, self.ny),
                frac_nan=FRAC_NAN,
            ),
            coords={
                "member": members,
                "init": inits,
                "lon": lons,
                "lat": lats,
                "lead": leads,
            },
            dims=("member", "init", "lead", "lon", "lat"),
            name="var",
            attrs={"units": "var units", "description": "a description"},
        )
        self.observations["var"] = xr.DataArray(
            randn((self.ninit, self.nx, self.ny), frac_nan=FRAC_NAN),
            coords={"lon": lons, "lat": lats, "time": inits},
            dims=("time", "lon", "lat"),
            name="var",
            attrs={"units": "var units", "description": "a description"},
        )

        self.uninit["var"] = xr.DataArray(
            randn((self.ninit, self.nx, self.ny, self.nmember), frac_nan=FRAC_NAN),
            coords={"lon": lons, "lat": lats, "time": inits, "member": members},
            dims=("time", "lon", "lat", "member"),
            name="var",
            attrs={"units": "var units", "description": "a description"},
        )

        self.hind.attrs = {"history": "created for xarray benchmarking"}
        self.hind.lead.attrs["units"] = "years"
        self.uninit.time.attrs["units"] = "years"
        self.observations.time.attrs["units"] = "years"


class Compute(Generate):
    """
    Benchmark time and peak memory of `compute_hindcast` and `bootstrap_hindcast`.
    """

    def setup(self, *args, **kwargs):
        self.make_hind_obs()

    @parameterized(["metric", "comparison"], (METRICS, HINDCAST_COMPARISONS))
    def time_compute_hindcast(self, metric, comparison):
        """Take time for `compute_hindcast`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else "init"
        ensure_loaded(
            compute_hindcast(
                self.hind,
                self.observations,
                metric=metric,
                comparison=comparison,
                dim=dim,
            )
        )

    @parameterized(["metric", "comparison"], (METRICS, HINDCAST_COMPARISONS))
    def peakmem_compute_hindcast(self, metric, comparison):
        """Take memory peak for `compute_hindcast`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else "init"
        ensure_loaded(
            compute_hindcast(
                self.hind,
                self.observations,
                metric=metric,
                comparison=comparison,
                dim=dim,
            )
        )

    @parameterized(["metric", "comparison"], (METRICS, HINDCAST_COMPARISONS))
    def time_bootstrap_hindcast(self, metric, comparison):
        """Take time for `bootstrap_hindcast`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else "init"
        ensure_loaded(
            bootstrap_hindcast(
                self.hind,
                self.uninit,
                self.observations,
                metric=metric,
                comparison=comparison,
                iterations=self.iterations,
                dim=dim,
            )
        )

    @parameterized(["metric", "comparison"], (METRICS, HINDCAST_COMPARISONS))
    def peakmem_bootstrap_hindcast(self, metric, comparison):
        """Take memory peak for `bootstrap_hindcast`."""
        dim = "member" if metric in PROBABILISTIC_METRICS else "init"
        ensure_loaded(
            bootstrap_hindcast(
                self.hind,
                self.uninit,
                self.observations,
                metric=metric,
                comparison=comparison,
                iterations=self.iterations,
                dim=dim,
            )
        )


class ComputeDask(Compute):
    def setup(self, *args, **kwargs):
        """Benchmark time and peak memory of `compute_hindcast` and
        `bootstrap_hindcast`. This executes the same tests as `Compute` but on chunked
        data."""
        requires_dask()
        # magic taken from
        # https://github.com/pydata/xarray/blob/stable/asv_bench/benchmarks/rolling.py
        super().setup(**kwargs)
        # chunk along a spatial dimension to enable embarrasingly parallel computation
        self.hind = self.hind["var"].chunk()
        self.observations = self.observations["var"].chunk()
        self.uninit = self.uninit["var"].chunk()


class ComputeDaskDistributed(ComputeDask):
    def setup(self, *args, **kwargs):
        """Benchmark time and peak memory of `compute_hindcast` and
        `bootstrap_hindcast`. This executes the same tests as `Compute` but
        on chunked data with dask.distributed.Client."""
        requires_dask()
        # magic taken from
        # https://github.com/pydata/xarray/blob/stable/asv_bench/benchmarks/rolling.py
        super().setup(**kwargs)
        self.client = Client()

    def cleanup(self):
        self.client.shutdown()


class ComputeSmall(Compute):
    def setup(self, *args, **kwargs):
        """Benchmark time and peak memory of `compute_hindcast` and
        `bootstrap_hindcast`. This executes the same tests as `Compute` but on 1D
        data."""
        requires_dask()
        # magic taken from
        # https://github.com/pydata/xarray/blob/stable/asv_bench/benchmarks/rolling.py
        super().setup(**kwargs)
        # chunk along a spatial dimension to enable embarrasingly parallel computation
        spatial_dims = ["lon", "lat"]
        self.hind = self.hind.mean(spatial_dims)
        self.observations = self.observations.mean(spatial_dims)
        self.uninit = self.uninit.mean(spatial_dims)
        self.iterations = 500



import xarray as xr

from .alignment import return_inits_and_verif_dates
from .checks import has_valid_lead_units, is_in_list, is_xarray
from .comparisons import (
    COMPARISON_ALIASES,
    HINDCAST_COMPARISONS,
    PM_COMPARISONS,
    PROBABILISTIC_HINDCAST_COMPARISONS,
    PROBABILISTIC_PM_COMPARISONS,
)
from .constants import CLIMPRED_DIMS, CONCAT_KWARGS, M2M_MEMBER_DIM, PM_CALENDAR_STR
from .exceptions import DimensionError
from .logging import log_compute_hindcast_header, log_compute_hindcast_inits_and_verifs
from .metrics import HINDCAST_METRICS, METRIC_ALIASES, PM_METRICS
from .reference import (
    _adapt_member_for_reference_forecast,
    climatology,
    persistence,
    uninitialized,
)
from .utils import (
    assign_attrs,
    convert_time_index,
    get_comparison_class,
    get_lead_cftime_shift_args,
    get_metric_class,
    shift_cftime_singular,
)


def _apply_metric_at_given_lead(
    verif,
    verif_dates,
    lead,
    hind=None,
    hist=None,
    inits=None,
    reference=None,
    metric=None,
    comparison=None,
    dim=None,
    **metric_kwargs,
):
    """Applies a metric between two time series at a given lead.

    Args:
        verif (xr object): Verification data.
        verif_dates (dict): Lead-dependent verification dates for alignment.
        lead (int): Given lead to score.
        hind (xr object): Initialized hindcast. Not required in a persistence forecast.
        hist (xr object): Uninitialized/historical simulation. Required when
            ``reference='uninitialized'``.
        inits (dict): Lead-dependent initialization dates for alignment.
        reference (str): If not ``None``, return score for this reference forecast.
            * 'persistence'
            * 'uninitialized'
        metric (Metric): Metric class for scoring.
        comparison (Comparison): Comparison class.
        dim (str): Dimension to apply metric over.

    Returns:
        result (xr object): Metric results for the given lead for the initialized
            forecast or reference forecast.
    """
    # naming:: lforecast: forecast at lead; lverif: verification at lead
    if reference is None:
        # Use `.where()` instead of `.sel()` to account for resampled inits when
        # bootstrapping.
        lforecast = hind.sel(lead=lead).where(hind["time"].isin(inits[lead]), drop=True)
        lverif = verif.sel(time=verif_dates[lead])
    elif reference == "persistence":
        lforecast, lverif = persistence(verif, inits, verif_dates, lead)
    elif reference == "uninitialized":
        lforecast, lverif = uninitialized(hist, verif, verif_dates, lead)
    elif reference == "climatology":
        lforecast, lverif = climatology(verif, inits, verif_dates, lead)
    if reference is not None:
        lforecast, dim = _adapt_member_for_reference_forecast(
            lforecast, lverif, metric, comparison, dim
        )

    lforecast["time"] = lverif[
        "time"
    ]  # a bit dangerous: what if different? more clear once https://github.com/pangeo-data/climpred/issues/523#issuecomment-728951645 implemented
    dim = _rename_dim(
        dim, hind, verif
    )  # dim should be much clearer once time in initialized.coords
    if metric.normalize or metric.allows_logical:
        metric_kwargs["comparison"] = comparison

    result = metric.function(lforecast, lverif, dim=dim, **metric_kwargs)
    log_compute_hindcast_inits_and_verifs(dim, lead, inits, verif_dates, reference)
    # push time (later renamed to init) back by lead
    if "time" in result.dims:
        n, freq = get_lead_cftime_shift_args(hind.lead.attrs["units"], lead)
        result = result.assign_coords(time=shift_cftime_singular(result.time, -n, freq))
    return result


def _rename_dim(dim, forecast, verif):
    """rename `dim` to `time` or `init` if forecast and verif dims require to do so."""
    if "init" in dim and "time" in forecast.dims and "time" in verif.dims:
        dim = dim.copy()
        dim.remove("init")
        dim = dim + ["time"]
    elif "time" in dim and "init" in forecast.dims and "init" in verif.dims:
        dim = dim.copy()
        dim.remove("time")
        dim = dim + ["init"]
    elif "init" in dim and "time" in forecast.dims and "time" in verif.dims:
        dim = dim.copy()
        dim.remove("init")
        dim = dim + ["time"]
    return dim


def _sanitize_to_list(dim):
    """Make dim to list if string, tuple or set, pass if None else raise ValueError."""
    if isinstance(dim, str):
        dim = [dim]
    elif isinstance(dim, set):
        dim = list(dim)
    elif isinstance(dim, tuple):
        dim = list(dim)
    elif isinstance(dim, list) or dim is None:
        pass
    else:
        raise ValueError(
            f"Expected `dim` as `str`, `list` or None, found {dim} as type {type(dim)}."
        )
    return dim


def _get_metric_comparison_dim(initialized, metric, comparison, dim, kind):
    """Returns `metric`, `comparison` and `dim` for compute functions.

    Args:
        initialized (xr.object): initialized dataset: init_pm or hind
        metric (str): metric or alias string
        comparison (str): Description of parameter `comparison`.
        dim (list of str or str): dimension to apply metric to.
        kind (str): experiment type from ['hindcast', 'PM'].

    Returns:
        metric (Metric): metric class
        comparison (Comparison): comparison class.
        dim (list of str or str): corrected dimension to apply metric to.
    """
    dim = _sanitize_to_list(dim)

    # check kind allowed
    is_in_list(kind, ["hindcast", "PM"], "kind")

    if dim is None:  # take all dimension from initialized except lead
        dim = list(initialized.dims)
        if "lead" in dim:
            dim.remove("lead")
        # adjust dim for e2c comparison when member not in forecast or verif
        if comparison in ["e2c"] and "member" in dim:
            dim.remove("member")

    # check that initialized contains all dims from dim
    if not set(dim).issubset(initialized.dims):
        raise DimensionError(
            f"`dim`={dim} is expected to be a subset of "
            f"`initialized.dims`={initialized.dims}."
        )

    # get metric and comparison strings incorporating alias
    metric = METRIC_ALIASES.get(metric, metric)
    comparison = COMPARISON_ALIASES.get(comparison, comparison)

    METRICS = HINDCAST_METRICS if kind == "hindcast" else PM_METRICS
    COMPARISONS = HINDCAST_COMPARISONS if kind == "hindcast" else PM_COMPARISONS
    metric = get_metric_class(metric, METRICS)
    comparison = get_comparison_class(comparison, COMPARISONS)

    # check whether combination of metric and comparison works
    PROBABILISTIC_COMPARISONS = (
        PROBABILISTIC_HINDCAST_COMPARISONS
        if kind == "hindcast"
        else PROBABILISTIC_PM_COMPARISONS
    )
    if metric.probabilistic:
        if not comparison.probabilistic and "member" in initialized.dims:
            raise ValueError(
                f"Probabilistic metric `{metric.name}` requires comparison "
                f"accepting multiple members e.g. `{PROBABILISTIC_COMPARISONS}`, "
                f"found `{comparison.name}`."
            )
        if "member" not in dim and "member" in initialized.dims:
            raise ValueError(
                f"Probabilistic metric {metric.name} requires to be "
                f"computed over dimension `member`, which is not found in {dim}."
            )
    else:  # determinstic metric
        if kind == "hindcast":
            # for thinking in real time as compute_hindcast renames init to time
            dim = ["time" if d == "init" else d for d in dim]
    return metric, comparison, dim


@is_xarray([0])
def compute_perfect_model(
    init_pm,
    control=None,
    metric="pearson_r",
    comparison="m2e",
    dim=["member", "init"],
    add_attrs=True,
    **metric_kwargs,
):
    """
    Compute a predictability skill score for a perfect-model framework
    simulation dataset.

    Args:
        init_pm (xarray object): ensemble with dims ``lead``, ``init``, ``member``.
        control (xarray object): NOTE that this is a legacy argument from a former
            release. ``control`` is not used in ``compute_perfect_model`` anymore.
        metric (str): `metric` name, see
         :py:func:`climpred.utils.get_metric_class` and (see :ref:`Metrics`).
        comparison (str): `comparison` name defines what to take as forecast
            and verification (see
            :py:func:`climpred.utils.get_comparison_class` and :ref:`Comparisons`).
        dim (str or list of str): dimension to apply metric over.
            default: ['member', 'init']
        add_attrs (bool): write climpred compute args to attrs. default: True
        ** metric_kwargs (dict): additional keywords to be passed to metric.
            (see the arguments required for a given metric in metrics.py)

    Returns:
        skill (xarray object): skill score with dimensions as input `ds`
                               without `dim`.

    """
    # Check that init is int, cftime, or datetime; convert ints or datetime to cftime
    init_pm = convert_time_index(
        init_pm, "init", "init_pm[init]", calendar=PM_CALENDAR_STR
    )

    # check args compatible with each other
    metric, comparison, dim = _get_metric_comparison_dim(
        init_pm, metric, comparison, dim, kind="PM"
    )

    forecast, verif = comparison.function(init_pm, metric=metric)

    if metric.normalize or metric.allows_logical:
        metric_kwargs["comparison"] = comparison
    skill = metric.function(forecast, verif, dim=dim, **metric_kwargs)
    if comparison.name == "m2m" and M2M_MEMBER_DIM in skill.dims:
        skill = skill.mean(M2M_MEMBER_DIM)
    # Attach climpred compute information to skill
    if add_attrs:
        skill = assign_attrs(
            skill,
            init_pm,
            metric=metric,
            comparison=comparison,
            dim=dim,
            metadata_dict=metric_kwargs,
        )
    return skill


@is_xarray([0, 1])
def compute_hindcast(
    hind,
    verif,
    metric="pearson_r",
    comparison="e2o",
    dim="init",
    alignment="same_verifs",
    add_attrs=True,
    **metric_kwargs,
):
    """Verify hindcast predictions against verification data.

    Args:
        hind (xarray object): Hindcast ensemble.
            Expected to follow package conventions:
            * ``init`` : dim of initialization dates
            * ``lead`` : dim of lead time from those initializations
            Additional dims can be member, lat, lon, depth, ...
        verif (xarray object): Verification data with some temporal overlap with the
            hindcast.
        metric (str): Metric used in comparing the decadal prediction ensemble with the
            verification data. (see :py:func:`~climpred.utils.get_metric_class` and
            :ref:`Metrics`).
        comparison (str):
            How to compare the decadal prediction ensemble to the verification data:

                * e2o : ensemble mean to verification data (Default)
                * m2o : each member to the verification data
                (see :ref:`Comparisons`)
        dim (str or list): dimension to apply metric over. default: 'init'
        alignment (str): which inits or verification times should be aligned?
            - maximize/None: maximize the degrees of freedom by slicing ``hind`` and
            ``verif`` to a common time frame at each lead.
            - same_inits: slice to a common init frame prior to computing
            metric. This philosophy follows the thought that each lead should be based
            on the same set of initializations.
            - same_verif: slice to a common/consistent verification time frame prior to
            computing metric. This philosophy follows the thought that each lead
            should be based on the same set of verification dates.
        add_attrs (bool): write climpred compute args to attrs. default: True
        **metric_kwargs (dict): additional keywords to be passed to metric
            (see the arguments required for a given metric in :ref:`Metrics`).

    Returns:
        result (xarray object):
            Verification metric over ``lead`` reduced by dimension(s) ``dim``.
    """
    metric, comparison, dim = _get_metric_comparison_dim(
        hind, metric, comparison, dim, kind="hindcast"
    )
    hind = convert_time_index(hind, "init", "hind[init]")
    verif = convert_time_index(verif, "time", "verif[time]")
    has_valid_lead_units(hind)

    forecast, verif = comparison.function(hind, verif, metric=metric)

    # think in real time dimension: real time = init + lag
    forecast = forecast.rename({"init": "time"})

    inits, verif_dates = return_inits_and_verif_dates(
        forecast, verif, alignment=alignment
    )

    if "iteration" in forecast.dims and "iteration" not in verif.dims:
        verif = verif.expand_dims(iteration=forecast.iteration)

    log_compute_hindcast_header(metric, comparison, dim, alignment, "initialized")

    metric_over_leads = [
        _apply_metric_at_given_lead(
            verif,
            verif_dates,
            lead,
            hind=forecast,
            inits=inits,
            metric=metric,
            comparison=comparison,
            dim=dim,
            **metric_kwargs,
        )
        for lead in forecast["lead"].data
    ]
    result = xr.concat(metric_over_leads, dim="lead", **CONCAT_KWARGS)
    result["lead"] = forecast["lead"]
    # rename back to 'init'
    if "time" in result.dims:
        result = result.rename({"time": "init"})
    # These computations sometimes drop coordinates along the way. This appends them
    # back onto the results of the metric.

    # Attach climpred compute information to result
    if add_attrs:
        result = assign_attrs(
            result,
            hind,
            alignment=alignment,
            metric=metric,
            comparison=comparison,
            dim=dim,
            metadata_dict=metric_kwargs,
        )
    return result



import numpy as np
import pytest

from climpred import HindcastEnsemble, PerfectModelEnsemble
from climpred.bootstrap import bootstrap_perfect_model
from climpred.checks import DimensionError
from climpred.graphics import plot_bootstrapped_skill_over_leadyear

ITERATIONS = 3


def test_mpi_he_plot_bootstrapped_skill_over_leadyear_da(
    PM_da_initialized_1d, PM_da_control_1d
):
    """
    Checks plots from bootstrap_perfect_model works for xr.DataArray.
    """
    res = bootstrap_perfect_model(
        PM_da_initialized_1d,
        PM_da_control_1d,
        metric="pearson_r",
        iterations=ITERATIONS,
        reference="uninitialized",
    )
    res_ax = plot_bootstrapped_skill_over_leadyear(res)
    assert res_ax is not None


def test_mpi_he_plot_bootstrapped_skill_over_leadyear_single_uninit_lead(
    PM_da_initialized_1d, PM_da_control_1d
):
    """
    Checks plots from bootstrap_perfect_model works for xr.DataArray.
    """
    res = bootstrap_perfect_model(
        PM_da_initialized_1d,
        PM_da_control_1d,
        metric="pearson_r",
        iterations=ITERATIONS,
        reference=["uninitialized", "persistence"],
    )
    # set all but first uninit lead to nan
    res[:, 2, 1:] = [np.nan] * (res.lead.size - 1)
    res_ax = plot_bootstrapped_skill_over_leadyear(res)
    assert res_ax is not None


def test_mpi_he_plot_bootstrapped_skill_over_leadyear_ds(
    PM_ds_initialized_1d, PM_ds_control_1d
):
    """
    Checks plots from bootstrap_perfect_model works for xr.Dataset with one variable.
    """
    res = bootstrap_perfect_model(
        PM_ds_initialized_1d,
        PM_ds_control_1d,
        metric="pearson_r",
        iterations=ITERATIONS,
        reference="uninitialized",
    )
    assert list(res.coords["skill"]) == ["initialized", "uninitialized"]
    res_ax = plot_bootstrapped_skill_over_leadyear(res)
    assert res_ax is not None


@pytest.mark.parametrize("cmap", ["tab10", "jet"])
@pytest.mark.parametrize("show_members", [True, False])
@pytest.mark.parametrize("variable", ["tos", None])
def test_PerfectModelEnsemble_plot(
    PM_ds_initialized_1d, PM_ds_control_1d, variable, show_members, cmap
):
    """Test PredictionEnsemble.plot()."""
    pm = PerfectModelEnsemble(PM_ds_initialized_1d)
    kws = {"cmap": cmap, "show_members": show_members, "variable": variable}
    pm.plot(**kws)
    pm = pm.add_control(PM_ds_control_1d)
    pm.plot(**kws)
    pm = pm.generate_uninitialized()
    pm.plot(**kws)


def test_PerfectModelEnsemble_plot_fails_3d(PM_ds_initialized_3d):
    """Test PredictionEnsemble.plot()."""
    pm = PerfectModelEnsemble(PM_ds_initialized_3d)
    with pytest.raises(DimensionError) as excinfo:
        pm.plot()
    assert "does not allow dimensions other" in str(excinfo.value)


@pytest.mark.parametrize("cmap", ["tab10", "jet"])
@pytest.mark.parametrize("show_members", [True, False])
@pytest.mark.parametrize("variable", ["SST", None])
def test_HindcastEnsemble_plot(
    hind_ds_initialized_1d,
    hist_ds_uninitialized_1d,
    reconstruction_ds_1d,
    observations_ds_1d,
    variable,
    show_members,
    cmap,
):
    """Test PredictionEnsemble.plot()."""
    he = HindcastEnsemble(hind_ds_initialized_1d)
    kws = {"cmap": cmap, "show_members": show_members, "variable": variable}
    he.plot(**kws)
    he = he.add_uninitialized(hist_ds_uninitialized_1d)
    he.plot(**kws)
    he = he.add_observations(reconstruction_ds_1d)
    he.plot(**kws)
    he = he.add_observations(observations_ds_1d)
    he.plot(**kws)



import warnings

import numpy as np
import xarray as xr

try:
    from eofs.xarray import Eof
except ImportError:
    Eof = None


def _relative_entropy_formula(sigma_b, sigma_x, mu_x, mu_b, neofs):
    """
    Compute the relative entropy formula given in Branstator and Teng, (2010).

    Args:
        sigma_b (xr.DataArray): covariance matrix of baseline distribution
        sigma_x (xr.DataArray): covariance matrix of forecast distribution
        mu_b (xr.DataArray): mean state vector of the baseline distribution
        mu_x (xr.DataArray): mean state vector of the forecast distribution
        neofs (int): number of EOFs used

    Returns:
        R (float): relative entropy
        dispersion (float): dispersion component
        signal (float): signal component

    References:
        * Branstator, Grant, and Haiyan Teng. “Two Limits of Initial-Value
          Decadal Predictability in a CGCM.” Journal of Climate 23, no. 23
          (August 27, 2010): 6292–6311. https://doi.org/10/bwq92h.
        * Kleeman, Richard. “Measuring Dynamical Prediction Utility Using
          Relative Entropy.” Journal of the Atmospheric Sciences 59, no. 13
          (July 1, 2002): 2057–72. https://doi.org/10/fqwxpk.

    """
    fac = 0.5
    dispersion = fac * (
        np.log(np.linalg.det(sigma_b) / np.linalg.det(sigma_x))
        + np.trace(sigma_x / sigma_b)
        - neofs
    )
    # https://stackoverflow.com/questions/
    # 7160162/left-matrix-division-and-numpy-solve
    # (A.T\B)*A
    x, _, _, _ = np.linalg.lstsq(
        sigma_b, mu_x - mu_b, rcond=None
    )  # sigma_b \ (mu_x - mu_b)
    signal = fac * np.matmul((mu_x.values - mu_b.values), x)
    R = dispersion + signal
    return R, dispersion, signal


def _bootstrap_dim(control, nlead_years, dim, dim_label):
    """
    Add a `len(dim_label)` dimension `dim` to uninitialized control by random
    resampling.
    """
    c_start = 0
    c_end = control["time"].size
    leads = np.arange(1, 1 + nlead_years)

    def isel_years(control, year_s, length=nlead_years):
        new = control.isel(time=slice(year_s, year_s + length))
        new = new.rename({"time": "lead"})
        new["lead"] = leads
        return new

    def create_pseudo_members(control):
        startlist = np.random.randint(c_start, c_end - nlead_years - 1, len(dim_label))
        return xr.concat([isel_years(control, start) for start in startlist], dim)

    control_uninitialized = create_pseudo_members(control)
    control_uninitialized[dim] = dim_label
    return control_uninitialized


# TODO: refactoring needed. proposed steps:
# first calculate all EOFs. save those. then calc compute_relative_entropy
def compute_relative_entropy(
    initialized,
    control,
    anomaly_data=False,
    neofs=None,
    curv=True,
    nlead=None,
    nmember_control=10,
):
    """
    Compute relative entropy.

    Calculates EOFs from anomalies. Projects fields on EOFs to receive
    pseudo-Principle Components per init and lead year. Calculate
    relative entropy based on _relative_entropy_formula.

    Args:
        initialized (xr.Dataset): anomaly ensemble data with dimensions
                                    lead, member, time and
                                    spatial [lon (x), lat(y)].
                                    DPLE or PM_ds
        control (xr.Dataset): anomaly control distribution with
                                              non-spatial dimensions:
                                              spatial [lon (x), lat(y)].
                                              - LENS: member, time
                                              - PM_control: time
        anomaly_data (bool): Input data is anomaly alread. Default: False.
        neofs (int): number of EOFs to use.
                     Default: initialized.member.size.
        curv (bool): if curvilinear grids disables EOF weights.
        nlead (int): number of timesteps calculated.
        nmember_control (int): number of members created from
                               bootstrapping from control

    Returns:
        rel_ent (xr.Dataset): relative entropy
    """
    if Eof is None:
        raise ImportError(
            "eofs is not installed; see"
            "https://ajdawson.github.io/eofs/latest/index.html"
        )
    # Defaults
    if neofs is None:
        neofs = initialized.member.size
    if nlead is None:
        nlead = initialized.lead.size

    # case if you submit control with dim time and member, LENS case
    if "member" in control.dims:
        control_uninitialized = _bootstrap_dim(
            control,
            initialized.lead.size,
            dim="init",
            dim_label=list(initialized.init.values),
        )

    # case if you only submit control with dim time, PM case
    else:
        control_uninitialized = xr.concat(
            [
                _bootstrap_dim(
                    control,
                    initialized.lead.size,
                    dim="member",
                    dim_label=np.arange(nmember_control),
                )
                for _ in range(initialized.init.size)
            ],
            dim="init",
        )
        control_uninitialized["init"] = initialized.init.values

    # initialized and control_uninitialized are allowed to have different
    # dims as I need more members to sample my control distr. properly
    if set(initialized.dims) != set(control_uninitialized.dims):
        warnings.warn(
            "Warning: initialized and control_uninitialized have different coords."
        )
        # print(initialized, control_uninitialized)

    # convert to xr.Data.Array
    if isinstance(control_uninitialized, xr.Dataset):
        control_uninitialized = control_uninitialized.to_array().squeeze()
    if isinstance(initialized, xr.Dataset):
        initialized = initialized.to_array().squeeze()

    # detrend
    non_spatial_dims = set(control_uninitialized.dims).intersection(["init", "member"])
    non_spatial_dims = list(non_spatial_dims)
    if not anomaly_data:  # if ds, control are raw values
        anom_x = initialized - control_uninitialized.mean(non_spatial_dims)
        anom_b = control_uninitialized - control_uninitialized.mean(non_spatial_dims)
    else:  # leave as is when already anomalies
        anom_x = initialized
        anom_b = control_uninitialized

    # prepare for EOF
    if curv:  # if curvilinear lon(x,y), lat(x,y) data inputs
        wgts = None
    else:  # assumes there is 'lat' in coords
        coslat = np.cos(np.deg2rad(anom_x.coords["lat"].values))
        wgts = np.sqrt(coslat)[..., np.newaxis]

    # EOF requires xr.dataArray
    if isinstance(control, xr.Dataset):
        control = control.to_array().squeeze()

    if "member" in control.dims:  # LENS
        # stack member and init into time dim, make time first
        non_spatial_control_dims = list(
            set(control.dims).intersection(["time", "member"])
        )

        transpose_dims = list(control.dims)
        transpose_dims.remove("member")
        transpose_dims.remove("time")
        dims = tuple(["time"] + transpose_dims)

        base_to_calc_eofs = (
            control.stack(new=tuple(non_spatial_control_dims))
            .rename({"new": "time"})
            .set_index({"time": "time"})
            .transpose(*dims)
        )
    else:
        # PM_control
        base_to_calc_eofs = control

    solver = Eof(base_to_calc_eofs, weights=wgts)

    re_leadtime_list = []
    leads = initialized.lead.values[:nlead]
    inits = initialized.init.values
    # DoTo: parallelize this double loop
    for init in inits:  # loop over inits
        rl, sl, dl = ([] for _ in range(3))  # lists to store results in
        for lead in leads:  # loop over lead time
            # P_b base distribution # eofs require time
            pc_b = solver.projectField(
                anom_b.sel(init=init, lead=lead)
                .drop("lead")
                .rename({"member": "time"}),
                neofs=neofs,
                eofscaling=0,
                weighted=False,
            ).rename({"time": "lead"})

            mu_b = pc_b.mean("lead")
            sigma_b = xr.DataArray(np.cov(pc_b.T))

            # P_x init distribution
            pc_x = solver.projectField(
                anom_x.sel(init=init, lead=lead)
                .drop("lead")
                .rename({"member": "time"}),
                neofs=neofs,
                eofscaling=0,
                weighted=False,
            ).rename({"time": "lead"})

            mu_x = pc_x.mean("lead")
            sigma_x = xr.DataArray(np.cov(pc_x.T))

            r, d, s = _relative_entropy_formula(sigma_b, sigma_x, mu_x, mu_b, neofs)

            rl.append(r)
            sl.append(s)
            dl.append(d)

        re_leadtime_list.append(
            xr.Dataset({"R": ("lead", rl), "S": ("lead", sl), "D": ("lead", dl)})
        )

    re = xr.concat(re_leadtime_list, dim="init").assign(init=inits, lead=leads)

    return re


def bootstrap_relative_entropy(
    initialized,
    control,
    sig=95,
    bootstrap=100,
    curv=True,
    neofs=None,
    nlead=None,
    anomaly_data=False,
    nmember_control=15,
):
    """
    Bootstrap relative entropy threshold.

    Generates a random uninitialized init and calculates the relative
    entropy. sig-th percentile determines threshold level.

    Args:
        initialized (xr.DataArray): initialized ensemble with dimensions
                                    init, member, time, lon (x),
                                    lat(y).
        control_uninitialized (xr.DataArray): control distribution with
                                              dimensions time,
                                              lon (x), lat(y).
        sig (int): significance level for threshold.
        bootstrap (int): number of bootstrapping iterations.
        neofs (int): number of EOFs to use. Default: initialized.member.size
        nlead (int): number of lead timestep to calculate.
                     Default: initialized.lead.size.
        curv (bool): if curvilinear grids are provided disables EOF weights.

    Returns:
        rel_ent (pd.DataFrame): relative entropy sig-th percentile threshold.

    """
    if neofs is None:
        neofs = initialized.member.size
    if nlead is None:
        nlead = initialized.lead.size
    if bootstrap < nlead:
        bootstrap = nlead + 1

    def _create_uninitialized_ensemble_from_control(ds, control, member_label):
        """Create uninitialized ensemble from control."""
        control_uninitialized = xr.concat(
            [
                _bootstrap_dim(
                    control, ds.lead.size, dim="member", dim_label=member_label
                )
                for _ in range(ds.init.size)
            ],
            dim="init",
        )
        control_uninitialized["init"] = ds.init.values
        return control_uninitialized

    results_list = []
    for _ in range(min(1, int(bootstrap / initialized.lead.size))):
        if "member" in control.dims:  # resample from lens
            uninitialized_initialized = _bootstrap_dim(
                control,
                initialized.lead.size,
                dim="init",
                dim_label=initialized.init.values,
            )
        else:  # PM
            uninitialized_initialized = _create_uninitialized_ensemble_from_control(
                initialized, control, list(initialized.member.values)
            )
        ds_pseudo_rel_ent = compute_relative_entropy(
            uninitialized_initialized,
            control,
            neofs=neofs,
            curv=curv,
            nlead=nlead,
            anomaly_data=anomaly_data,
            nmember_control=nmember_control,
        )
        results_list.append(ds_pseudo_rel_ent)
    ds_pseudo_metric = xr.concat(results_list, dim="bootstrap")
    qsig = sig / 100
    sig_level = ds_pseudo_metric.quantile(q=qsig, dim=["bootstrap", "lead", "init"])
    return sig_level



import numpy as np
import xarray as xr

from .checks import is_in_list
from .constants import VALID_ALIGNMENTS
from .exceptions import CoordinateError
from .utils import get_multiple_lead_cftime_shift_args, shift_cftime_index


def return_inits_and_verif_dates(forecast, verif, alignment, reference=None, hist=None):
    """Returns initializations and verification dates for an arbitrary number of leads
    per a given alignment strategy.

    Args:
        forecast (``xarray`` object): Prediction ensemble with ``init`` dim renamed to
            ``time`` and containing ``lead`` dim.
        verif (``xarray`` object): Verification data with ``time`` dim.
        alignment (str): Strategy for initialization-verification alignment.
            * 'same_inits': Use a common set of initializations that verify
               across all leads. This ensures that there is no bias in the result due
               to the state of the system for the given initializations.
            * 'same_verifs': Use a common verification window across all leads. This
               ensures that there is no bias in the result due to the observational
               period being verified against.
            * 'maximize': Use all available initializations at each lead that verify
               against the observations provided. This changes both the set of
               initializations and the verification window used at each lead.

    Returns:
        inits (dict): Keys are the lead time integer, values are an ``xr.DataArray`` of
            initialization dates.
        verif_dates (dict): Keys are the lead time integer, values are an
            ``xr.CFTimeIndex`` of verification dates.
    """
    if isinstance(reference, str):
        reference = [reference]
    elif reference is None:
        reference = []

    is_in_list(alignment, VALID_ALIGNMENTS, "alignment")
    units = forecast["lead"].attrs["units"]
    leads = forecast["lead"].values

    # `init` renamed to `time` in compute functions.
    all_inits = forecast["time"]
    all_verifs = verif["time"]

    # If aligning reference='uninitialized', need to account for potential differences
    # in its temporal coverage. Note that the reference='uninitialized' only aligns
    # verification dates and doesn't care about inits.
    if hist is not None:
        all_verifs = np.sort(list(set(all_verifs.data) & set(hist["time"].data)))
        all_verifs = xr.DataArray(all_verifs, dims=["time"], coords=[all_verifs])

    # Construct list of `n` offset over all leads.
    n, freq = get_multiple_lead_cftime_shift_args(units, leads)
    init_lead_matrix = _construct_init_lead_matrix(forecast, n, freq, leads)

    # A union between `inits` and observations in the verification data is required
    # for persistence, since the persistence forecast is based off a common set of
    # initializations.
    if "persistence" in reference:
        union_with_verifs = all_inits.isin(all_verifs)
        init_lead_matrix = init_lead_matrix.where(union_with_verifs, drop=True)
    valid_inits = init_lead_matrix["time"]

    if "same_init" in alignment:
        return _same_inits_alignment(
            init_lead_matrix, valid_inits, all_verifs, leads, n, freq
        )
    elif "same_verif" in alignment:
        return _same_verifs_alignment(
            init_lead_matrix, valid_inits, all_verifs, leads, n, freq
        )
    elif alignment == "maximize":
        return _maximize_alignment(init_lead_matrix, all_verifs, leads)


def _maximize_alignment(init_lead_matrix, all_verifs, leads):
    """Returns initializations and verification dates, maximizing the degrees of freedom
    at each lead individually.

    See ``return_inits_and_verif_dates`` for descriptions of expected variables.
    """
    # Move row-wise and find all forecasted times that align with verification dates at
    # the given lead.
    verify_with_observations = init_lead_matrix.isin(all_verifs)
    lead_dependent_verif_dates = init_lead_matrix.where(verify_with_observations)
    # Probably a way to do this more efficiently since we're doing essentially
    # the same thing at each step.
    verif_dates = {
        lead: lead_dependent_verif_dates.sel(lead=lead).dropna("time").to_index()
        for lead in leads
    }
    inits = {
        lead: lead_dependent_verif_dates.sel(lead=lead).dropna("time")["time"]
        for lead in leads
    }
    return inits, verif_dates


def _same_inits_alignment(init_lead_matrix, valid_inits, all_verifs, leads, n, freq):
    """Returns initializations and verification dates, maintaining a common set of inits
    at all leads.

    See ``return_inits_and_verif_dates`` for descriptions of expected variables.
    """
    verifies_at_all_leads = init_lead_matrix.isin(all_verifs).all("lead")
    inits = valid_inits.where(verifies_at_all_leads, drop=True)
    inits = {lead: inits for lead in leads}
    verif_dates = {
        lead: shift_cftime_index(inits[lead], "time", n, freq)
        for (lead, n) in zip(leads, n)
    }
    return inits, verif_dates


def _same_verifs_alignment(init_lead_matrix, valid_inits, all_verifs, leads, n, freq):
    """Returns initializations and verification dates, maintaining a common verification
    window at all leads.

    See ``return_inits_and_verif_dates`` for descriptions of expected variables.
    """
    common_set_of_verifs = [
        i for i in all_verifs if (i == init_lead_matrix).any("time").all("lead")
    ]
    if not common_set_of_verifs:
        raise CoordinateError(
            "A common set of verification dates cannot be found for the "
            "initializations and verification data supplied. Change `alignment` to "
            "'same_inits' or 'maximize'."
        )
    # Force to CFTimeIndex for consistency with `same_inits`
    verif_dates = xr.concat(common_set_of_verifs, "time").to_index()
    inits_that_verify_with_verif_dates = init_lead_matrix.isin(verif_dates)
    inits = {
        lead: valid_inits.where(
            inits_that_verify_with_verif_dates.sel(lead=lead), drop=True
        )
        for lead in leads
    }
    verif_dates = {lead: verif_dates for lead in leads}
    return inits, verif_dates


def _construct_init_lead_matrix(forecast, n, freq, leads):
    """Returns xr.DataArray of "real time" (init + lead) over all inits and leads.

    Arguments:
        forecast (``xarray object``): Prediction ensemble with ``init`` dim renamed to
            ``time`` and containing ``lead`` dim.
        n (tuple of ints): Number of units to shift for ``leads``. ``value`` for
            ``CFTimeIndex.shift(value, str)``.
        freq (str): Pandas frequency alias. ``str`` for
            ``CFTimeIndex.shift(value, str)``.
        leads (list, array, xr.DataArray of ints): Leads to return offset for.

    Returns:
        init_lead_matrix (``xr.DataArray``): DataArray with x=inits and y=lead with
            values corresponding to "real time", or ``init + lead`` over all inits and
            leads.
    """
    # Note that `init` is renamed to `time` in compute functions.
    init_lead_matrix = xr.concat(
        [
            xr.DataArray(
                shift_cftime_index(forecast, "time", n, freq),
                dims=["time"],
                coords=[forecast["time"]],
            )
            for n in n
        ],
        "lead",
    )
    init_lead_matrix["lead"] = leads
    return init_lead_matrix



import dask
import numpy as np
import pytest
import xarray as xr

from climpred.bootstrap import (
    _bootstrap_by_stacking,
    _chunk_before_resample_iterations_idx,
    _resample,
    _resample_iterations,
    _resample_iterations_idx,
    bootstrap_hindcast,
    bootstrap_uninit_pm_ensemble_from_control_cftime,
)
from climpred.comparisons import HINDCAST_COMPARISONS
from climpred.constants import CONCAT_KWARGS, VALID_ALIGNMENTS
from climpred.exceptions import KeywordError
from climpred.utils import _transpose_and_rechunk_to

# TODO: move to conftest.py
ITERATIONS = 2

comparison_dim_PM = [
    ("m2m", "init"),
    ("m2m", "member"),
    ("m2m", ["init", "member"]),
    ("m2e", "init"),
    ("m2e", "member"),
    ("m2e", ["init", "member"]),
    ("m2c", "init"),
    ("m2c", "member"),
    ("m2c", ["init", "member"]),
    ("e2c", "init"),
]

xr.set_options(display_style="text")


def test_bootstrap_PM_keep_lead_attrs(perfectModelEnsemble_initialized_control):
    """Test bootstrap_perfect_model works lazily."""
    pm = perfectModelEnsemble_initialized_control
    pm.get_initialized().lead.attrs["units"] = "years"
    s = pm.bootstrap(
        iterations=ITERATIONS,
        comparison="m2c",
        metric="mse",
    )
    assert "units" in s.lead.attrs
    assert s.lead.attrs["units"] == pm.get_initialized().lead.attrs["units"]


@pytest.mark.parametrize("comparison,dim", comparison_dim_PM)
@pytest.mark.parametrize("chunk", [True, False])
def test_bootstrap_PM_lazy_results(
    perfectModelEnsemble_initialized_control, chunk, comparison, dim
):
    """Test bootstrap_perfect_model works lazily."""
    pm = perfectModelEnsemble_initialized_control.isel(lead=range(3))
    if chunk:
        pm = pm.chunk({"lead": 2}).chunk({"time": -1})
    else:
        pm = pm.compute()
    s = pm.bootstrap(
        iterations=ITERATIONS,
        comparison=comparison,
        metric="mse",
        dim=dim,
    )
    assert dask.is_dask_collection(s) == chunk


@pytest.mark.slow
@pytest.mark.parametrize("chunk", [True, False])
def test_bootstrap_hindcast_lazy(
    hindcast_hist_obs_1d,
    chunk,
):
    """Test bootstrap_hindcast works lazily."""
    he = hindcast_hist_obs_1d.isel(lead=range(3), init=range(10))
    if chunk:
        he = he.chunk({"lead": 2})
    else:
        he = he.compute()

    s = he.bootstrap(
        iterations=ITERATIONS,
        comparison="e2o",
        metric="mse",
        alignment="same_verifs",
        dim="init",
    )
    assert dask.is_dask_collection(s) == chunk


@pytest.mark.slow
@pytest.mark.parametrize("resample_dim", ["member", "init"])
def test_bootstrap_hindcast_resample_dim(
    hindcast_hist_obs_1d,
    resample_dim,
):
    """Test bootstrap_hindcast when resampling member or init and alignment
    same_inits."""
    hindcast_hist_obs_1d.isel(lead=range(3), init=range(10)).bootstrap(
        iterations=ITERATIONS,
        comparison="e2o",
        metric="mse",
        resample_dim=resample_dim,
        alignment="same_inits",
        dim="init",
    )


def bootstrap_uninit_pm_ensemble_from_control(init_pm, control):
    """
    Create a pseudo-ensemble from control run. Deprecated in favor of
    `bootstrap_uninit_pm_ensemble_from_control_cftime`.

    Note:
        Needed for block bootstrapping confidence intervals of a metric in perfect
        model framework. Takes randomly segments of length of ensemble dataset from
        control and rearranges them into ensemble and member dimensions.

    Args:
        init_pm (xarray object): ensemble simulation.
        control (xarray object): control simulation.

    Returns:
        uninit (xarray object): pseudo-ensemble generated from control run.
    """
    nens = init_pm.init.size
    nmember = init_pm.member.size
    length = init_pm.lead.size
    c_start = 0
    c_end = control["time"].size
    lead_time = init_pm["lead"]

    def set_coords(uninit, init, dim):
        uninit[dim] = init[dim].values
        return uninit

    def isel_years(control, year_s, length):
        new = control.isel(time=slice(year_s, year_s + length))
        new = new.rename({"time": "lead"})
        new["lead"] = lead_time
        return new

    def create_pseudo_members(control):
        startlist = np.random.randint(c_start, c_end - length - 1, nmember)
        uninit_ens = xr.concat(
            (isel_years(control, start, length) for start in startlist),
            dim="member",
            **CONCAT_KWARGS,
        )
        return uninit_ens

    uninit = xr.concat(
        (
            set_coords(create_pseudo_members(control), init_pm, "member")
            for _ in range(nens)
        ),
        dim="init",
        **CONCAT_KWARGS,
    )
    # chunk to same dims
    return (
        _transpose_and_rechunk_to(uninit, init_pm)
        if dask.is_dask_collection(uninit)
        else uninit
    )


def test_bootstrap_uninit_pm_ensemble_from_control_cftime_annual_identical(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
):
    """Test ``bootstrap_uninit_pm_ensemble_from_control_cftime`` cftime identical to
    ``bootstrap_uninit_pm_ensemble_from_control`` for annual data."""
    cftime_res = bootstrap_uninit_pm_ensemble_from_control_cftime(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )

    noncftime_res = bootstrap_uninit_pm_ensemble_from_control(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )
    # lead and member identical
    for d in ["lead", "member"]:
        assert (cftime_res[d] == noncftime_res[d]).all()
    # init same size
    assert cftime_res["init"].size == noncftime_res["init"].size
    assert cftime_res.dims == noncftime_res.dims
    assert list(cftime_res.data_vars) == list(noncftime_res.data_vars)


def test_bootstrap_uninit_pm_ensemble_from_control_cftime_annual_identical_da(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
):
    """Test ``bootstrap_uninit_pm_ensemble_from_control_cftime`` cftime identical to
    ``bootstrap_uninit_pm_ensemble_from_control`` for annual data."""
    PM_ds_initialized_1d_ym_cftime = PM_ds_initialized_1d_ym_cftime["tos"]
    PM_ds_control_1d_ym_cftime = PM_ds_control_1d_ym_cftime["tos"]
    cftime_res = bootstrap_uninit_pm_ensemble_from_control_cftime(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )
    noncftime_res = bootstrap_uninit_pm_ensemble_from_control(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )
    # lead and member identical
    for d in ["lead", "member"]:
        assert (cftime_res[d] == noncftime_res[d]).all()
    # init same size
    assert cftime_res["init"].size == noncftime_res["init"].size
    assert cftime_res.name == noncftime_res.name
    # assert cftime_res.shape == noncftime_res.shape
    # assert cftime_res.dims == noncftime_res.dims


@pytest.mark.parametrize(
    "init, control",
    [
        (
            pytest.lazy_fixture("PM_ds_initialized_1d_ym_cftime"),
            pytest.lazy_fixture("PM_ds_control_1d_ym_cftime"),
        ),
        (
            pytest.lazy_fixture("PM_ds_initialized_1d_mm_cftime"),
            pytest.lazy_fixture("PM_ds_control_1d_mm_cftime"),
        ),
        (
            pytest.lazy_fixture("PM_ds_initialized_1d_dm_cftime"),
            pytest.lazy_fixture("PM_ds_control_1d_dm_cftime"),
        ),
    ],
)
def test_bootstrap_uninit_pm_ensemble_from_control_cftime_all_freq(init, control):
    """Test bootstrap_uninit_pm_ensemble_from_control_cftime for all freq data."""
    init = init.isel(lead=range(3), init=range(5))
    uninit = bootstrap_uninit_pm_ensemble_from_control_cftime(init, control)
    # lead and member identical
    for d in ["lead", "member"]:
        assert (uninit[d] == init[d]).all()
    # init same size
    assert uninit["init"].size == init["init"].size


def test_bootstrap_by_stacking_dataset(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
):
    res = _bootstrap_by_stacking(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )
    assert res.lead.attrs["units"] == "years"
    assert isinstance(res, xr.Dataset)
    assert res.tos.dims == PM_ds_initialized_1d_ym_cftime.tos.dims


def test_bootstrap_by_stacking_dataarray(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
):
    v = list(PM_ds_initialized_1d_ym_cftime.data_vars)[0]
    res = _bootstrap_by_stacking(
        PM_ds_initialized_1d_ym_cftime[v], PM_ds_control_1d_ym_cftime[v]
    )
    assert res.lead.attrs["units"] == "years"
    assert isinstance(res, xr.DataArray)
    assert res.dims == PM_ds_initialized_1d_ym_cftime[v].dims


def test_bootstrap_by_stacking_chunked(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
):
    res_chunked = _bootstrap_by_stacking(
        PM_ds_initialized_1d_ym_cftime.chunk(),
        PM_ds_control_1d_ym_cftime.chunk(),
    )
    assert dask.is_dask_collection(res_chunked)
    res_chunked = res_chunked.compute()
    res = _bootstrap_by_stacking(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )
    for d in ["lead", "member"]:
        assert (res_chunked[d] == res[d]).all()
    # init same size
    assert res_chunked["init"].size == res["init"].size


def test_bootstrap_by_stacking_two_var_dataset(
    PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
):
    """Test _bootstrap_by_stacking when init_pm and control two variable dataset."""
    PM_ds_initialized_1d_ym_cftime["sos"] = PM_ds_initialized_1d_ym_cftime["tos"]
    PM_ds_control_1d_ym_cftime["sos"] = PM_ds_control_1d_ym_cftime["tos"]
    res = _bootstrap_by_stacking(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )
    res_cf = bootstrap_uninit_pm_ensemble_from_control_cftime(
        PM_ds_initialized_1d_ym_cftime, PM_ds_control_1d_ym_cftime
    )
    assert len(list(res.data_vars)) == len(list(res_cf.data_vars))
    # lead and member identical
    for d in ["lead", "member"]:
        assert (res[d] == res_cf[d]).all()
    # init same size
    assert res["init"].size == res_cf["init"].size


def test_bootstrap_hindcast_raises_error(
    hind_da_initialized_1d, hist_da_uninitialized_1d, observations_da_1d
):
    """Test that error is raised when user tries to resample over init and align over
    same_verifs."""
    with pytest.raises(KeywordError):
        bootstrap_hindcast(
            hind_da_initialized_1d,
            hist_da_uninitialized_1d,
            observations_da_1d,
            iterations=ITERATIONS,
            comparison="e2o",
            metric="mse",
            resample_dim="init",
            alignment="same_verifs",
        )


def test_resample_1_size(PM_da_initialized_1d):
    """Tests that the resampled dimensions are appropriate for a single iteration."""
    dim = "member"
    expected = _resample(PM_da_initialized_1d, resample_dim=dim)
    # 1 somehow fails
    actual = _resample_iterations_idx(PM_da_initialized_1d, 2, dim=dim).isel(
        iteration=0
    )
    assert expected.size == actual.size
    assert expected[dim].size == actual[dim].size


def test_resample_size(PM_da_initialized_1d):
    """Tests that the resampled dimensions are appropriate for many iterations."""
    dim = "member"
    expected = xr.concat(
        [_resample(PM_da_initialized_1d, resample_dim=dim) for i in range(ITERATIONS)],
        "iteration",
    )
    actual = _resample_iterations_idx(PM_da_initialized_1d, ITERATIONS, dim=dim)
    assert expected.size == actual.size
    assert expected[dim].size == actual[dim].size


@pytest.mark.parametrize("chunk", [True, False])
@pytest.mark.parametrize("replace", [True, False])
def test_resample_iterations_same(PM_da_initialized_1d, chunk, replace):
    """Test that both `resample_iterations` functions yield same result shape."""
    ds = PM_da_initialized_1d.isel(lead=range(3), init=range(5))
    if chunk:
        ds = ds.chunk()
    ds_r_idx = _resample_iterations_idx(ds, ITERATIONS, "member", replace=replace)
    ds_r = _resample_iterations(ds, ITERATIONS, "member", replace=replace)
    for d in ds.dims:
        xr.testing.assert_identical(ds_r[d], ds_r_idx[d])
        assert ds_r.size == ds_r_idx.size


def test_chunk_before_resample_iterations_idx(PM_da_initialized_3d_full):
    """Test that chunksize after `_resample_iteration_idx` is lower than
    `optimal_blocksize`."""
    chunking_dims = ["x", "y"]
    iterations = 50
    optimal_blocksize = 100000000
    ds_chunked = _chunk_before_resample_iterations_idx(
        PM_da_initialized_3d_full.chunk(),
        iterations,
        chunking_dims,
        optimal_blocksize=optimal_blocksize,
    )
    ds_chunked_chunksize = ds_chunked.data.nbytes / ds_chunked.data.npartitions
    print(
        dask.utils.format_bytes(ds_chunked_chunksize * iterations),
        "<",
        dask.utils.format_bytes(1.5 * optimal_blocksize),
    )
    assert ds_chunked_chunksize * iterations < 1.5 * optimal_blocksize


@pytest.mark.parametrize("chunk", [True, False])
@pytest.mark.parametrize("replace", [True, False])
def test_resample_iterations_dim_max(PM_da_initialized_1d, chunk, replace):
    """Test that both `resample_iterations(dim_max=n)` gives n members."""
    ds = PM_da_initialized_1d.isel(lead=range(3), init=range(5))
    ds = ds.sel(member=list(ds.member.values) * 2)
    ds["member"] = np.arange(1, 1 + ds.member.size)
    if chunk:
        ds = ds.chunk()
    ds_r = _resample_iterations(
        ds,
        ITERATIONS,
        "member",
        replace=replace,
        dim_max=PM_da_initialized_1d.member.size,
    )
    assert (ds_r["member"] == PM_da_initialized_1d.member).all()


@pytest.mark.skip(reason="this is a bug, test fails and should be resolved.")
def test_resample_iterations_dix_no_squeeze(PM_da_initialized_1d):
    """Test _resample_iteration_idx with singular dimension.

    Currently this fails for dimensions with just a single index as we use `squeeze` in
    the code and not using squeeze doesnt maintain functionality. This means that
    _resample_iteration_idx should not be called on singleton dimension inputs (which
    is not critical and can be circumvented when using squeeze before climpred.).
    """
    da = PM_da_initialized_1d.expand_dims("test_dim")
    print(da)
    actual = _resample_iterations_idx(da, iterations=ITERATIONS)
    assert "test_dim" in actual.dims


@pytest.mark.parametrize("metric", ["acc", "mae"])
def test_bootstrap_p_climatology(hindcast_hist_obs_1d, metric):
    """Test that p from bootstrap is close to 0 if skillful."""
    reference = "climatology"
    bskill = hindcast_hist_obs_1d.bootstrap(
        metric=metric,
        comparison="e2o",
        dim="init",
        iterations=21,
        alignment="same_inits",
        reference=reference,
    )
    v = "SST"
    lead = 1
    # first lead skill full
    if metric in ["acc"]:
        assert (
            (
                bskill.sel(skill="initialized", results="verify skill")
                > bskill.sel(skill=reference, results="verify skill")
            )
            .sel(lead=lead)[v]
            .all()
        )
    else:
        assert (
            (
                bskill.sel(skill="initialized", results="verify skill")
                < bskill.sel(skill=reference, results="verify skill")
            )
            .sel(lead=lead)[v]
            .all()
        )
    assert bskill.sel(skill=reference, results="p").sel(lead=lead)[v] < 0.1
